import builtins as _builtins
from importlib.metadata import version
from . import config
from . import main
from .main import (
    radius_search,
    radius_sum,
    radius_count,
    radius_mean,
    radius_variance,
    radius_std,
    radius_cv,
    radius_skewness,
    radius_kurtosis,
    detect_cluster_pts,
    detect_cluster_cells,
    detect_cluster_cells_from_labeled_pts,
)
from .radius_search.grid_class import Grid
from .radius_search.sample_area import infer_sample_area_from_pts
from .radius_search.null_distribution import draw_random_coords
from .utils.grid_aggregate import aggregate_to_grid

__version__ = version('aabpl')
if not getattr(_builtins, '_aabpl_imported', False):
    print(f"aabpl v{__version__}.")
    _builtins._aabpl_imported = True



from numpy import (
    array as _np_array,
    linspace as _np_linspace,
    stack as _np_stack,
    arange as _np_arange,
    unique as _np_unique,
    zeros as _np_zeros,
    min as _np_min,
    max as _np_max,
    mean as _np_mean,
    median as _np_median,
    std as _np_std,
    ceil as _np_ceil,
    pi as _np_pi,
    sum as _np_sum,
    log2 as _np_log2,
)
from pyproj import Transformer
from pandas import DataFrame as _pd_DataFrame
from math import log10 as _math_log10, inf as _math_inf
from aabpl.utils.misc import flatten_list, find_column_name
from aabpl.utils.crs_transformation import convert_bounds_to_local_crs
from aabpl.illustrations.plot_utils import map_2D_to_rgb, get_2D_rgb_colobar_kwargs
from .disk_search_state import (
    aggregate_point_data_to_cells,
    search_and_aggregate
)
from .point_region_assignment import assign_points_to_cell_regions
from .sample_area import compute_disk_cell_overlap, intersect_polygon_with_grid
from aabpl.testing.test_performance import time_func_perf
# from .clusters import (
#     create_clusters, add_geom_to_cluster, connect_cells_to_clusters,
#     make_cluster_orthogonally_convex, make_cluster_convex, merge_clusters,
#     add_cluster_tags_to_cells, save_full_grid, save_sparse_grid, save_cell_clusters)
from .clusters import Clustering
from shapely.geometry import Polygon, Point
from shapely.ops import unary_union
from geopandas import GeoDataFrame as _gpd_GeoDataFrame
# from decimal import Decimal as _decimal_Decimal, getcontext as _decimal_getcontext
import math
from .spacing_topology import (
    compute_spatial_stats, compute_spacing_breakpoints,
    choose_nest_depth,
    predict_timing, choose_spacing_and_depth,
)
from aabpl import config as _cfg

# def get_spacing_decimal(xmin, xmax, n_steps, ndec = 20):
#     _decimal_getcontext(ndec)
#     return _decimal_Decimal(xmax-xmin)/_decimal_Decimal(n_steps)


class _DFList(list):
    """list of GeoDataFrames that also supports dict-style lookup by cluster column name.

    Backwards-compatible: existing code using ``dfs[0]``, ``for df in dfs``, ``len(dfs)``
    continues to work unchanged. New code can use ``dfs['employment_cluster_sum_750']``
    or iterate via ``dfs.keys()`` / ``dfs.items()``.
    """
    def __init__(self):
        super().__init__()
        self._by_column = {}

    def _append(self, key, df):
        self.append(df)
        self._by_column[key] = df

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._by_column[item]
        return super().__getitem__(item)

    def keys(self):
        return self._by_column.keys()

    def values(self):
        return self._by_column.values()

    def items(self):
        return self._by_column.items()


class Bounds(object):
    __slots__ = ('xmin', 'xmax', 'ymin', 'ymax', 'np_array_of_bounds') # use this syntax to save some memory. also only create vars that are really neccessary
    def __init__(self, xmin:float, xmax:float, ymin:float, ymax:float):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
    #
#


class SearchInternals:
    """Namespace for internal search-grid state. Not part of the public API."""
    pass



        

class Grid(object):
    """
    Result object returned by ``detect_cluster_pts`` and ``detect_cluster_cells``.

    All coordinates and geometries are in ``proj_crs`` (a projected CRS, typically UTM).

    CRS & bounds
    ------------
    input_crs : str
        Original CRS of the input ``pts`` DataFrame as passed by the user.
    proj_crs : str
        Projected CRS used for all internal coordinates and output geometries
        (e.g. ``'EPSG:32618'``).

    Output grid
    -----------
    cell_size : float
        Output cell size (default ``r / 3``). All display, export, and cluster
        geometry use this cell size.
    x_steps : numpy.ndarray
        Left-edge x-coordinates of output columns (set after the search runs).
    y_steps : numpy.ndarray
        Left-edge y-coordinates of output rows.
    cell_aggregates : dict
        ``{(row, col): values_array}`` — aggregated indicator values per
        non-empty output cell.  Keyed by integer output-grid row/col indices.
    cell_row_name : str
        Name of the output row-index column written to the input ``pts`` DataFrame.
    cell_col_name : str
        Name of the output col-index column written to the input ``pts`` DataFrame.

    Cell helpers
    ------------
    grid.cell_centroid(row, col) -> (x, y)
        Centre coordinate of output cell ``(row, col)`` in ``proj_crs``.
    grid.cell_bounds(row, col) -> (xmin, ymin, xmax, ymax)
        Bounding box of output cell ``(row, col)``.
    grid.cell_poly(row, col) -> [(x,y), ...]
        Four corner coordinates of output cell ``(row, col)``.

    Clustering
    ----------
    clustering : Clustering
        Cluster results, populated by ``detect_cluster_cells``.

        - ``grid.clustering.by_column`` — ``{col_name: ClustersForColumn}``
        - Each ``ClustersForColumn.clusters`` is a list of ``Cluster`` objects:

          - ``.id``       — integer cluster id (1-based)
          - ``.cells``    — set of ``(row, col)`` search-grid cell indices
          - ``.geometry`` — Shapely polygon/multipolygon in ``proj_crs``
          - ``.centroid`` — ``(x, y)`` in ``proj_crs``

    Plots
    -----
    plot : GridPlots
        All methods return a ``matplotlib.Figure``.

        - ``grid.plot.clusters()``    — cluster polygons over aggregated output grid
        - ``grid.plot.vars()``        — choropleth of radius-sum values per output cell
        - ``grid.plot.cluster_pts()`` — scatter coloured by radius-sum, clusters highlighted
        - ``grid.plot.rand_dist()``   — observed vs. null-distribution radius-sums

    Null distribution
    -----------------
    null_distribution : pandas.DataFrame
        Random points drawn for the null distribution, with their x/y coordinates
        and computed radius-sum columns (same names as the indicator columns plus
        the search suffix).  ``None`` if no null distribution was computed.
    sample_area : shapely.Polygon or MultiPolygon, optional
        Valid sampling area used to draw null-distribution random points
        (in ``proj_crs``).  ``None`` if no sample area was set.

    Export
    ------
    grid.clustering.create_clusters_df_for_column(cluster_column)
        GeoDataFrame with one row per cluster polygon.
    grid.clustering.save_cell_clusters(filename, file_format='shp', target_crs='initial')
        Save cluster polygons to shapefile or CSV.
    grid.clustering.create_sparse_grid_df(target_crs='initial')
        GeoDataFrame of non-empty output cells with aggregated values and cluster ids.
    grid.clustering.save_sparse_grid(filename, file_format='shp', target_crs='initial')
        Save the sparse grid to shapefile or CSV.

    Internals
    ---------
    _search_internals : SearchInternals
        Namespace bundling internal search-grid state (spacing, steps, bounds,
        codec, etc.).  Not part of the public API.
    _search_class : DiskSearch
        The active search state object.  Not part of the public API.
    """
    @time_func_perf
    def __init__(
        self,
        xmin:float,
        xmax:float,
        ymin:float,
        ymax:float,
        initial_crs:str,
        r:float,
        local_crs:str='auto',
        n_pts_src:int=None,
        n_pts_tgt:int=None,
        n_pts_src_extra:int=0,
        pts_tgt_xy=None,
        data_crs:str=None,
        output_spacing:float=None,
        output_spacing_y:float=None,
        silent=False,
    ):
        """
        Internal constructor — call via ``detect_cluster_pts`` / ``detect_cluster_cells``.

        The public-facing cell size is ``spacing`` (default ``r/3``).  An internal
        search grid with a different cell size is chosen automatically for performance;
        all internal search-grid state lives in ``_search_internals`` and is not part
        of the public API.
        """
        self._silent = silent
        # if local crs should be found automatically or
        if local_crs == 'auto' or initial_crs != local_crs:
            local_crs, (xmin,xmax,ymin,ymax) = convert_bounds_to_local_crs(
                xmin=xmin,
                xmax=xmax,
                ymin=ymin,
                ymax=ymax,
                initial_crs=initial_crs,
                target_crs=local_crs,
                silent=silent
            )
        # data_crs accepted but not stored (was never read publicly)

        # auto choose spacing ratio and depth unless explictly set by the user via config.
        spacing, nest_depth = choose_spacing_and_depth(
            r=r,
            spacing_ratio=_cfg.FIXED_SPACING_RATIO,
            nest_depth=_cfg.FIXED_NEST_DEPTH,
            n_pts_src=n_pts_src,
            n_pts_tgt=n_pts_tgt,
            n_pts_src_extra=n_pts_src_extra,
            pts_tgt_xy=pts_tgt_xy,
            silent=silent,
        )

        self._search_spacing = spacing
        self.nest_depth = nest_depth

        self.clustering = Clustering(self)
        from aabpl.illustrations.plot_grid import GridPlots
        self.plot = GridPlots(self)
        # TODO _search_internals.bounds should also contain excluded area if not cntd
        # min(points._search_internals.bounds+r, max(points._search_internals.bounds, excluded_area_total_bound))
        self.input_crs = initial_crs
        # Keep legacy aliases so old code that reads grid.initial_crs / grid.local_crs still works.
        self.initial_crs = initial_crs
        self.proj_crs = local_crs
        self.local_crs = local_crs
        x_padding = ((xmin-xmax) % spacing)/2
        y_padding = ((ymin-ymax) % spacing)/2
        _sb = Bounds(xmin=xmin-x_padding,xmax=xmax+x_padding,ymin=ymin-y_padding,ymax=ymax+y_padding)
        self.total_bounds = _sb  # legacy alias; canonical: grid._search_internals.bounds
        n_xsteps = -int((xmin-xmax)/spacing)+2 # round up
        n_ysteps = -int((ymin-ymax)/spacing)+2 # round up
        x_steps = _np_linspace(_sb.xmin, _sb.xmax, n_xsteps)
        y_steps = _np_linspace(_sb.ymin, _sb.ymax, n_ysteps)
        self._search_x_steps = x_steps
        self._search_y_steps = y_steps
        self._search_row_ids = _np_arange(n_ysteps-1)
        self._search_col_ids = _np_arange(n_xsteps-1)
        self._search_n_cells = len(self._search_row_ids)*len(self._search_col_ids)
        # Output grid step arrays and ids — initialised to None here; set by update_spacing().
        # row_ids / col_ids / n_cells / x_steps / y_steps point to the OUTPUT grid.
        self.x_steps = None
        self.y_steps = None
        self.row_ids = self._search_row_ids
        self.col_ids = self._search_col_ids
        self.n_cells = self._search_n_cells

        # Integer cell-key codec: packs (lvl, row, col) into one int64 so the
        # cell-sum aggregation and the per-search offset templates share the same
        # packing and a template translates to a point with one vector add. Sized
        # from the search-grid extent. Always built — the search loop relies on it.
        from aabpl.utils.cell_keys import CellKeyCodec
        self.cell_codec = CellKeyCodec(
            nest_depth=nest_depth,
            row_lo=int(self._search_row_ids.min()), row_hi=int(self._search_row_ids.max()),
            col_lo=int(self._search_col_ids.min()), col_hi=int(self._search_col_ids.max()),
            offset_margin=16,
        )
       
        # Output grid — user-facing cell size for exports and plots.
        # When unset, defaults to r/3: the search grid is chosen for speed (a coarser
        # cell, typically ~0.35r-0.71r), but the output raster should be fine enough to
        # render results crisply. Per-point aggregates are exact regardless of the
        # search cell size, so a finer output grid is well-defined.
        # Output spacing (the user-facing cell size) is fixed at construction, default r/3.
        # The output grid itself (arrays and cell aggregates) is built LAZILY by
        # update_spacing() — radius_search alone skips it to avoid the per-point
        # aggregation overhead when no output grid is needed.
        self.output_spacing = output_spacing if output_spacing is not None else r / 3
        self.output_spacing_y = output_spacing_y if output_spacing_y is not None else self.output_spacing
        # cell_size = output spacing (not search spacing). Set once here permanently.
        self.cell_size = self.output_spacing
        self.output_x_steps = None
        self.output_y_steps = None
        self.cell_aggregates = {}
        # Legacy alias kept for backwards compatibility.
        self.output_id_to_sums = self.cell_aggregates
        self._output_val_cols = []
        self._spacing_computed = False
        # Bundle all search-grid internals into a namespace for convenience.
        si = SearchInternals()
        si.spacing    = spacing
        si.x_steps    = x_steps
        si.y_steps    = y_steps
        si.row_ids    = self._search_row_ids
        si.col_ids    = self._search_col_ids
        si.n_cells    = self._search_n_cells
        si.nest_depth = nest_depth
        si.cell_codec = self.cell_codec
        si.raw_bounds = (xmin, xmax, ymin, ymax)
        si.bounds     = _sb
        self._search_internals = si
        # Legacy aliases kept for backwards compatibility.
        self._raw_bounds = si.raw_bounds

        self.sample_area = None
        self.sample_grid_bounds = None
        self.cells_rndm_sample = set()


        # TODO replace row_col_to_centroid with function fetching it from 1d arrays to no clog up memory for fine grids
        
        self.centroids = None
        self.row_col_to_centroid = None

        grid_xmin = _sb.xmin
        grid_ymin = _sb.ymin
        grid_xmax = _sb.xmax
        grid_ymax = _sb.ymax
        # self.get_cell_centroid = lambda row, col: (float(grid_xmin+col*(spacing+.5)),float(grid_ymin+row*(spacing+.5)))
        min_row, max_row, min_col, max_col = min(self._search_row_ids), max(self._search_row_ids), min(self._search_row_ids), max(self._search_col_ids)
        self._search_cell_centroid = lambda row, col, x_steps=self._search_x_steps, y_steps=self._search_y_steps: (
            float(
                x_steps[int(col):int(col)+2].sum()/2 if min_col<=col<=max_col else 
                grid_xmin+(col+.5)*spacing if min_col<col else 
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(row):int(row)+2].sum()/2 if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else 
                grid_ymax+(row-max_row-1+.5)*spacing
            )
            ) if row%1==0 and col%1==0 else (
            float(
                x_steps[int(round(col)):int(round(col))+2].sum()/2 + (col-round(col))*spacing if min_col<=col<=max_col else
                grid_xmin+(col+.5)*spacing if min_col<col else
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(round(row)):int(round(row))+2].sum()/2 + (row-round(row))*spacing if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else
                grid_ymax+(row-max_row-1+.5)*spacing
            )
            )
        # Legacy alias for existing code that calls grid.get_cell_centroid (search-grid coords).
        self.get_cell_centroid = self._search_cell_centroid

        self._search_cell_poly = lambda row, col, x_steps=self._search_x_steps, y_steps=self._search_y_steps: [
            ((x_mean-spacing/(2**(1+lvl)),y_mean-spacing/(2**(1+lvl))), 
             (x_mean+spacing/(2**(1+lvl)),y_mean-spacing/(2**(1+lvl))), 
             (x_mean+spacing/(2**(1+lvl)),y_mean+spacing/(2**(1+lvl))), 
             (x_mean-spacing/(2**(1+lvl)),y_mean+spacing/(2**(1+lvl))))
            for x_mean,y_mean,lvl in [(
            float(
                x_steps[int(col):int(col)+2].sum()/2 if min_col<=col<=max_col else 
                grid_xmin+(col+.5)*spacing if min_col<col else 
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(row):int(row)+2].sum()/2 if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else 
                grid_ymax+(row-max_row-1+.5)*spacing
            ),
            0
            ) if row%1==0 and col%1==0 else (
            float(
                x_steps[int(round(col)):int(round(col))+2].sum()/2 + (col-round(col))*spacing if min_col<=col<=max_col else 
                grid_xmin+(col+.5)*spacing if min_col<col else 
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(round(row)):int(round(row))+2].sum()/2 + (row-round(row))*spacing if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else 
                grid_ymax+(row-max_row-1+.5)*spacing
            ),
            next((i for i,n in enumerate(range(max(20,nest_depth+1))) if row%.5%(2**-(n+1))==0),max(21,nest_depth+2)) # UPDATE if max_nest_level > 20
            )]][0]
        
        # Legacy alias for existing code that calls grid.get_cell_poly.
        self.get_cell_poly = self._search_cell_poly

        self._search_cell_bounds = lambda row, col, x_steps=self._search_x_steps, y_steps=self._search_y_steps: [
            ((x_mean-spacing/(2**(1+lvl)),
              y_mean-spacing/(2**(1+lvl))), 
             (x_mean+spacing/(2**(1+lvl)),
              y_mean+spacing/(2**(1+lvl))))
            for x_mean,y_mean,lvl in [(
            float(
                x_steps[int(col):int(col)+2].sum()/2 if min_col<=col<=max_col else 
                grid_xmin+(col+.5)*spacing if min_col<col else 
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(row):int(row)+2].sum()/2 if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else 
                grid_ymax+(row-max_row-1+.5)*spacing
            ),
            0
            ) if row%1==0 and col%1==0 else (
            float(
                x_steps[int(round(col)):int(round(col))+2].sum()/2 + (col-round(col))*spacing if min_col<=col<=max_col else 
                grid_xmin+(col+.5)*spacing if min_col<col else 
                grid_xmax+(col-max_col-1+.5)*spacing
            ),
            float(
                y_steps[int(round(row)):int(round(row))+2].sum()/2 + (row-round(row))*spacing if min_row<=row<=max_row else
                grid_ymin+(row+.5)*spacing if min_row<row else 
                grid_ymax+(row-max_row-1+.5)*spacing
            ),
            next((i for i,n in enumerate(range(max(20,nest_depth+1))) if row%.5%(2**-(n+1))==0),max(21,nest_depth+2)) # UPDATE if max_nest_level > 20
            )]][0]
        # Legacy alias for existing code that calls grid.get_cell_bounds.
        self.get_cell_bounds = self._search_cell_bounds

        # NOTE: no grid-creation print here. The internal search grid is not
        # user-facing; the informative print happens in update_spacing() when the
        # OUTPUT grid is actually (re)built, so it fires once per spacing value.
        #
    #
    # ── Public output-cell helpers ────────────────────────────────────────────
    def cell_centroid(self, row, col):
        """Return the (x, y) centroid of output cell (row, col) in proj_crs."""
        return (
            self._search_internals.bounds.xmin + (col + 0.5) * self.cell_size,
            self._search_internals.bounds.ymin + (row + 0.5) * self.output_spacing_y,
        )

    def cell_bounds(self, row, col):
        """Return (x0, y0, x1, y1) bounds of output cell (row, col) in proj_crs."""
        x0 = self._search_internals.bounds.xmin + col * self.cell_size
        y0 = self._search_internals.bounds.ymin + row * self.output_spacing_y
        return (x0, y0, x0 + self.cell_size, y0 + self.output_spacing_y)

    def cell_poly(self, row, col):
        """Return corner list [(x0,y0),(x1,y0),(x1,y1),(x0,y1)] for output cell."""
        x0, y0, x1, y1 = self.cell_bounds(row, col)
        return [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]

    # add functions
    aggregate_point_data_to_cells = aggregate_point_data_to_cells
    assign_points_to_cell_regions = assign_points_to_cell_regions # OUTDATED? REPLACE WITH assign_points_to_mirco_regions?
    intersect_polygon_with_grid = intersect_polygon_with_grid
    search_and_aggregate = search_and_aggregate
    compute_disk_cell_overlap = compute_disk_cell_overlap
    
    def get_all_ids(
            self,
            store:bool=False,
    ):
        if hasattr(self,'ids') and not self.ids is None:
            return self.ids 
        col_ids = self.col_ids
        row_ids = self.row_ids
        ids = tuple(flatten_list([[(int(row_id), int(col_id)) for col_id in col_ids] for row_id in row_ids]))
        if store:
            self.ids = ids
        return ids
    #

    def get_all_centroids(
                self,
                store:bool=False,
        ):
            if not self.centroids is None:
                return self.centroids
            row_ids = self.row_ids
            col_ids = self.col_ids
            x_steps = self.x_steps
            y_steps = self.y_steps
            centroids = _np_array([centroid for centroid in flatten_list([
                [(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean()) for col_id in col_ids] 
                for row_id in row_ids]
            )])
            if store:
                self.centroids = centroids
            return centroids
        # self.bounds = flatten_list([
        #         [((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1])) for col_id in col_ids] 
        #         for row_id in row_ids])
        # TODO replace row_col_to_centroid with function fetching it from 1d arrays to no clog up memory for fine grids
        
    def get_all_row_col_to_centroids(
            self,
            store:bool=False,
        ):
        
        if not self.row_col_to_centroid is None:
            return self.row_col_to_centroid
        row_ids = self.row_ids
        col_ids = self.col_ids
        x_steps = self.x_steps
        y_steps = self.y_steps
        row_col_to_centroid = {g_row_col:centroid for (g_row_col,centroid) in flatten_list([
            [((row_id,col_id),(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean())) for col_id in col_ids] 
            for row_id in row_ids]
            )}
        if store:
            self.row_col_to_centroid = row_col_to_centroid
        return row_col_to_centroid


    def assign_output_cell_ids(
        self,
        pts: _pd_DataFrame,
        x: str,
        y: str,
        row_name: str,
        col_name: str,
    ) -> None:
        """
        Assign each point to its output grid cell and store the column names
        on the grid as ``self.cell_row_name`` / ``self.cell_col_name``
        (also available as ``self.output_row_name`` / ``self.output_col_name``).

        When output spacing equals internal spacing the existing row/col
        columns are reused (no new columns written to pts).  Otherwise new
        columns ``'out_{row_name}'`` / ``'out_{col_name}'`` are added to pts.
        """
        from numpy import floor as _np_floor
        # row/col columns are assigned on the search grid, so compare against it.
        if self.output_spacing == self._search_spacing and self.output_spacing_y == self._search_spacing:
            self.cell_row_name = self.output_row_name = row_name
            self.cell_col_name = self.output_col_name = col_name
            return
        out_col_name = 'out_' + col_name
        out_row_name = 'out_' + row_name
        pts[out_col_name] = _np_floor(
            (pts[x].values - self.output_x_steps[0]) / self.output_spacing
        ).astype(int)
        pts[out_row_name] = _np_floor(
            (pts[y].values - self.output_y_steps[0]) / self.output_spacing_y
        ).astype(int)
        self.cell_row_name = self.output_row_name = out_row_name
        self.cell_col_name = self.output_col_name = out_col_name

    def aggregate_pts_to_output_cells(
        self,
        pts: _pd_DataFrame,
        val_cols: list,
        x: str = 'lon',
        y: str = 'lat',
        agg: str = 'sum',
    ) -> dict:
        """
        Aggregate per-point values into output grid cells and store in
        ``self.output_id_to_sums``.

        Uses ``self.output_spacing`` / ``self.output_spacing_y`` and
        ``self.output_x_steps`` / ``self.output_y_steps`` — all set in
        ``__init__``.

        Args:
            pts: DataFrame containing ``x``, ``y``, and all ``val_cols``.
            val_cols: Column names to aggregate (must already exist in ``pts``).
            x: X-coordinate column (must be in the same projection as the grid).
            y: Y-coordinate column.
            agg: Aggregation method — ``'sum'``, ``'mean'``, or ``'count'``.

        Returns:
            ``self.output_id_to_sums`` — dict mapping ``(row, col)`` to a
            numpy array of aggregated values (one entry per val_col).
        """
        from numpy import floor as _np_floor
        from pandas import DataFrame as _pd_DataFrame_local

        xmin_out = self.output_x_steps[0]
        ymin_out = self.output_y_steps[0]
        sx = self.output_spacing
        sy = self.output_spacing_y

        col_idx = _np_floor((pts[x].values - xmin_out) / sx).astype(int)
        row_idx = _np_floor((pts[y].values - ymin_out) / sy).astype(int)

        val_cols = list(val_cols)
        tmp = _pd_DataFrame_local({'_row': row_idx, '_col': col_idx})
        for col in val_cols:
            tmp[col] = pts[col].values

        grp = tmp.groupby(['_row', '_col'])[val_cols]
        if agg == 'sum':
            agg_df = grp.sum()
        elif agg == 'mean':
            agg_df = grp.mean()
        elif agg == 'count':
            agg_df = grp.count()
        else:
            raise ValueError(f"agg must be 'sum', 'mean', or 'count'; got {agg!r}")

        result = {rc: row.values for rc, row in agg_df.iterrows()}
        self.cell_aggregates = result
        self.output_id_to_sums = result  # legacy alias
        self._output_val_cols = val_cols
        return self.cell_aggregates

    def update_spacing(self, spacing: float = None, spacing_y: float = None, recompute: bool = False):
        """
        (Re)build the output grid and its cached cell aggregates.

        This is intentionally LAZY: it is NOT run when the grid is created, and
        ``radius_search`` does not call it — computing the output grid means a
        per-point aggregation that is wasted overhead if the caller only wants the
        radius search. It IS called by every consumer that needs the output grid:
        ``detect_cluster_pts`` / ``detect_cluster_cells`` (always), and the plot /
        export methods (``cell_aggregates``, ``clusters``, ``save_*``).

        What it (re)computes:
          - output grid arrays ``output_x_steps`` / ``output_y_steps``;
          - the public grid attributes ``cell_size`` / ``x_steps`` / ``y_steps`` /
            ``row_ids`` / ``col_ids`` / ``n_cells`` (re-pointed to the output grid);
          - the cached per-output-cell aggregate of the raw indicator columns in
            ``output_id_to_sums`` (read by ``cell_aggregates`` / ``clusters``);
          - the ``out_*`` cell-id columns on the target points.

        Parameters
        ----------
        spacing : float, optional
            New output cell size. If given and different, the output grid is rebuilt
            for it. Defaults to keeping the current ``output_spacing`` (r/3 unless set
            at construction).
        spacing_y : float, optional
            Output cell height; defaults to ``spacing``.
        recompute : bool
            Force recomputation even if already current.
        """
        changed = False
        if spacing is not None:
            new_y = spacing_y if spacing_y is not None else spacing
            if spacing != self.output_spacing or new_y != self.output_spacing_y:
                self.output_spacing = spacing
                self.output_spacing_y = new_y
                changed = True
        if self._spacing_computed and not changed and not recompute:
            return self

        xmin, xmax, ymin, ymax = self._search_internals.raw_bounds
        tb = self._search_internals.bounds
        n_out_xsteps = -int((xmin - xmax) / self.output_spacing) + 2
        n_out_ysteps = -int((ymin - ymax) / self.output_spacing_y) + 2
        self.output_x_steps = _np_linspace(tb.xmin, tb.xmax, n_out_xsteps)
        self.output_y_steps = _np_linspace(tb.ymin, tb.ymax, n_out_ysteps)
        # Re-point public grid attributes to the output grid. Search internals read
        # grid._search_*; the get_cell_* lambdas captured the search arrays at init.
        self.cell_size = self.output_spacing
        self.x_steps = self.output_x_steps
        self.y_steps = self.output_y_steps
        self.row_ids = _np_arange(len(self.output_y_steps) - 1)
        self.col_ids = _np_arange(len(self.output_x_steps) - 1)
        self.n_cells = len(self.row_ids) * len(self.col_ids)
        self.row_col_to_centroid = None  # centroid cache depends on the grid
        self.centroids = None

        # Cache per-output-cell aggregate of the RAW indicator columns (what
        # cell_aggregates / clusters display) and assign out_* cell ids.
        search = getattr(self, '_search_class', None) or getattr(self, 'search', None)
        tgt = getattr(search, 'target', None) if search is not None else None
        if tgt is not None and len(getattr(tgt, 'c', [])):
            self.aggregate_pts_to_output_cells(tgt.pts, val_cols=list(tgt.c), x=tgt.x, y=tgt.y, agg='sum')
            self.assign_output_cell_ids(tgt.pts, x=tgt.x, y=tgt.y, row_name=tgt.row_name, col_name=tgt.col_name)
        self._spacing_computed = True
        return self

    @time_func_perf
    def calc_micro_region_stats(self):
        """
        Calculate statistics on the micro regions created by the radius search, 
        such as the average number of distinct cntd and overlapped cells per region, 
        and their area. 
        Useful for understanding the characteristics of the micro regions and 
        for comparing different radius search configurations in optimization.
        """
        # Helper function to avoid redundant list comprehension syntax
        def calc_cells_area(cells):
            import numpy as _np_local
            if isinstance(cells, _np_local.ndarray):
                return float(_np_local.sum(2.0 ** (-2.0 * cells[:, 0]))) if len(cells) else 0.0
            return sum(2**(-2 * lvl) for lvl, _ in cells)

        _sc = getattr(self, '_search_class', None) or self.search
        # 1. Base shared metrics
        n_shared = {
            "cntd": len(_sc.shared_cntd_cells),
            "ovlpd": 0
        }
        area_shared = {
            "cntd": calc_cells_area(_sc.shared_cntd_cells),
            "ovlpd": 0
        }

        # 2. Accumulators for regions
        totals = {
            "count": {"cntd": 0.0, "ovlpd": 0.0},
            "area":  {"cntd": 0.0, "ovlpd": 0.0}
        }
        total_reg_area = 0.0
        total_reg_count = 0

        # 3. Process regions
        mult = _sc.region_and_trgl_mult
        for reg_id, reg_area in _sc.region_id_to_area.items():
            total_reg_area += reg_area
            total_reg_count += 1

            # Use triangle 1 as representative (all triangles share the same list unless
            # by_trgl overrides exist; for aggregate stats any triangle is equivalent)
            cells_data = {
                "cntd": _sc.region_and_trgl_id_to_distinct_cntd_cells[reg_id * mult + 1],
                "ovlpd": _sc.region_and_trgl_id_to_distinct_ovlpd_cells[reg_id * mult + 1]
            }

            for state in ["cntd", "ovlpd"]:
                cells = cells_data[state]
                c_count = len(cells)
                c_area = calc_cells_area(cells)

                # Accumulate values
                totals["count"][state] += c_count
                totals["area"][state]  += c_area
                
                # Weighted by region area
                totals["count"][f"{state}_weighted"] += c_count * reg_area
                totals["area"][f"{state}_weighted"]  += c_area * reg_area

        # 4. Normalization and final aggregation
        for state in ["cntd", "ovlpd"]:
            if total_reg_count > 0:
                # Normalize by count
                totals["count"][state] /= total_reg_count
                totals["area"][state]  /= total_reg_count
            
            if total_reg_area > 0:
                # Normalize by area
                totals["count"][f"{state}_weighted"] /= total_reg_area
                totals["area"][f"{state}_weighted"]  /= total_reg_area

            # Add shared components
            totals["count"][state] += n_shared[state]
            totals["area"][state]  += n_shared[state]

            totals["count"][f"{state}_weighted"] += area_shared[state]
            totals["area"][f"{state}_weighted"]  += area_shared[state]

        # 5. Normalise areas by circle area (π·r²), expressed in spacing² units.
        #    This ratio depends only on r/spacing and nest_depth, so it is
        #    reusable across different point clouds with the same geometry.
        from math import pi as _pi
        r = _sc.r
        spacing = self._search_spacing
        circle_area_in_spacing_units = _pi * (r / spacing) ** 2
        totals["area_share"] = {}
        for state in ["cntd", "ovlpd"]:
            for sfx in ["", "_weighted"]:
                key = state + sfx
                totals["area_share"][key] = (
                    totals["area"][key] / circle_area_in_spacing_units
                    if circle_area_in_spacing_units > 0 else 0.0
                )

        self.micro_region_stats = totals
        return totals
    
    # append plots
    # # append cluster functions
    # create_clusters = create_clusters
    # add_geom_to_cluster = add_geom_to_cluster
    # connect_cells_to_clusters = connect_cells_to_clusters
    # make_cluster_orthogonally_convex = make_cluster_orthogonally_convex
    # make_cluster_convex = make_cluster_convex
    # merge_clusters = merge_clusters
    # add_cluster_tags_to_cells = add_cluster_tags_to_cells
    # # # save options
    # save_full_grid = Clustering.save_full_grid
    # save_sparse_grid = Clustering.save_sparse_grid
    # save_cell_clusters = Clustering.save_cell_clusters

    def plot_sample_area(self,
        pts:_pd_DataFrame=None,
        x:str=None,
        y:str=None,
        filename:str='',
        plot_kwargs:dict={},
        show:bool=True,):
        from aabpl.illustrations.plot_sample_area import plot_sample_area
        plot_sample_area(
            grid=self,
            pts=pts or self.pts if hasattr(self,"pts") else None,
            x=x or self.x if hasattr(self,"x") else None,
            y=y or self.y if hasattr(self,"y") else None,
            filename=filename,
            plot_kwargs=plot_kwargs,
            show=show,
        )
    def create_full_grid_df(self, target_crs:str=['initial','local','EPSG:4326'][0], max_column_name_length:int=10):
        """returns geopandas.GeoDataFrame with entry for each grid cell with attributes on its Polygon, centroid, sum of indicator(s), and cluster id
        
        Args:
        -------
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        max_column_name_length (int):
            maximum length of automatically chosen target name (shapefiles allow a maximum column name length of 10)
        
        Returns:
        -------
        df (geopandas.GeoDataFrame):
            with entry for each grid cell
        """
        self.update_spacing()  # ensure the output grid is materialised
        c_ids = _np_zeros((self.n_cells, len(self.clustering.by_column)),int)#-1
        sums = _np_zeros((self.n_cells, len(self.clustering.by_column)),int)
        polys = []
        id_to_sums = self.id_to_sums
        centroids = _np_array(list(self.get_all_row_col_to_centroids().values()))
        target_crs = self.input_crs if target_crs=='initial' else self.proj_crs if target_crs=='local' else target_crs
        if target_crs != self.proj_crs:
            transformer = Transformer.from_crs(crs_from=self.proj_crs, crs_to=self.input_crs, always_xy=True)
            centroids_x, centroids_y = transformer.transform(centroids[:,0], centroids[:,1])
        else:
            centroids_x, centroids_y = centroids[:,0], centroids[:,1]
        clusters_for_columns = list(self.clustering.by_column.values())
        _sc = getattr(self, '_search_class', None) or self.search
        for (i, row_col), ((xmin,ymin),(xmax,ymax)) in zip(enumerate(self.ids), list(self.row_col_to_bounds.values())):
            for clusters_for_column in clusters_for_columns:
                if row_col in clusters_for_column.cell_to_cluster_id:
                    c_ids[i] = clusters_for_column.cell_to_cluster_id[row_col]
            if row_col in id_to_sums:
                sums[i] = id_to_sums[row_col]
            polys.append(Polygon(((xmin,ymin),(xmax,ymin),(xmax,ymax),(xmin,ymax))))
        df = _gpd_GeoDataFrame({
            _sc.source.row_name: [row for row,col in self.ids],
            _sc.source.col_name: [col for row,col in self.ids],
            'centroid_x': centroids_x,
            'centroid_y': centroids_y,
            }, geometry=polys,
            crs=self.proj_crs
            )
        if len(self.clustering.by_column)<=1:
            df['cluster_id'] = c_ids
            df['sum'] = sums
        else:
            for j, column in enumerate(self.clustering.by_column):
                c_id_colname = find_column_name("cluster_id", column, df.columns, max_column_name_length)
                agg_colname = find_column_name("sum_radius", column, df.columns, max_column_name_length)
                df[c_id_colname] = c_ids[:,j]
                df[agg_colname] = sums[:,j]
        if target_crs != self.proj_crs:
            df.to_crs(self.input_crs, inplace=True)

        return df
    #

    def create_sparse_grid_df(
            self, target_crs:str=['initial','local','EPSG:4326'][0], max_column_name_length:int=10
        ):
        """
        returns geopandas.GeoDataFrame with entry for grid cells that either has points inside or is part of a cluster with attributes on their Polygon, centroid, sum of indicator(s), and cluster id
        
        Args:
        -------
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        max_column_name_length (int):
            maximum length of automatically chosen target name (shapefiles allow a maximum column name length of 10)
        
        Returns:
        -------
        df (geopandas.GeoDataFrame):
            with entry for each grid cell
        """
        self.update_spacing()  # ensure the output grid + aggregates are materialised
        id_to_sums = self.output_id_to_sums
        x_steps = self.output_x_steps
        y_steps = self.output_y_steps
        out_row_ids = _np_arange(len(y_steps) - 1)
        out_col_ids = _np_arange(len(x_steps) - 1)
        n_out_cells = len(out_row_ids) * len(out_col_ids)
        polys = []
        target_crs = self.input_crs if target_crs=='initial' else self.proj_crs if target_crs=='local' else target_crs
        if target_crs != self.proj_crs:
            transformer = Transformer.from_crs(crs_from=self.proj_crs, crs_to=target_crs, always_xy=True)
        centroids_x = _np_zeros(n_out_cells, float)
        centroids_y = _np_zeros(n_out_cells, float)
        sparse_row_ids = _np_zeros(n_out_cells, int)
        sparse_col_ids = _np_zeros(n_out_cells, int)
        n_val_cols = max(len(self.clustering.by_column), 1)
        sums = _np_zeros((n_out_cells, n_val_cols), float)
        c_ids = _np_zeros((n_out_cells, len(self.clustering.by_column)), int)
        i = 0
        js_clusters_for_columns = list(enumerate(self.clustering.by_column.values()))
        for row in out_row_ids:
            ymin_c, ymax_c = y_steps[row], y_steps[row + 1]
            sparse_row_ids[i:] = row
            for col in out_col_ids:
                cell_in_a_cluster = False
                for j, clusters_for_column in js_clusters_for_columns:
                    if (row, col) in clusters_for_column.cell_to_cluster_id:
                        c_ids[i, j] = clusters_for_column.cell_to_cluster_id[(row, col)]
                        cell_in_a_cluster = True
                if (row, col) in id_to_sums:
                    vals = id_to_sums[(row, col)]
                    sums[i, :len(vals)] = vals
                elif not cell_in_a_cluster:
                    continue
                sparse_col_ids[i] = col
                xmin_c, xmax_c = x_steps[col], x_steps[col + 1]
                cx, cy = (xmin_c + xmax_c) / 2, (ymin_c + ymax_c) / 2
                if target_crs != self.proj_crs:
                    cx, cy = transformer.transform(cx, cy)
                centroids_x[i] = cx
                centroids_y[i] = cy
                polys.append(Polygon(((xmin_c, ymin_c), (xmax_c, ymin_c), (xmax_c, ymax_c), (xmin_c, ymax_c))))
                i += 1
        _sc2 = getattr(self, '_search_class', None) or getattr(self, 'search', None)
        out_row_name = getattr(self, 'output_row_name', _sc2.source.row_name if _sc2 else 'id_y')
        out_col_name = getattr(self, 'output_col_name', _sc2.source.col_name if _sc2 else 'id_x')
        df = _gpd_GeoDataFrame({
            out_row_name: sparse_row_ids[:i],
            out_col_name: sparse_col_ids[:i],
            'centroid_x': centroids_x[:i],
            'centroid_y': centroids_y[:i],
            }, geometry=polys,
            crs=self.proj_crs
            )
        val_cols = getattr(self, '_output_val_cols', [])
        if len(self.clustering.by_column) <= 1:
            df['cluster_id'] = c_ids[:i, 0] if c_ids.shape[1] > 0 else 0
            for k, vcol in enumerate(val_cols):
                df[vcol] = sums[:i, k]
        else:
            for j, column in enumerate(self.clustering.by_column):
                c_id_colname = find_column_name("cluster_id", column, df.columns, max_column_name_length)
                df[c_id_colname] = c_ids[:i, j]
            for k, vcol in enumerate(val_cols):
                df[vcol] = sums[:i, k]

        if target_crs != self.proj_crs:
            df.to_crs(target_crs, inplace=True)

        return df
    #

    def create_clusters_df_for_column(self, cluster_column:str, target_crs:str=['initial','local','EPSG:4326'][0]):
        """
        returns geopandas.GeoDataFrame with entry for grid cells that either has points inside or is part of a cluster with attributes on their Polygon, centroid, sum of indicator(s), and cluster id
        
        Args:
        cluster_column (str):
            column containing the variable by which the clustering shall be performed
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        
        Returns:
        df (geopandas.GeoDataFrame):
            with entry for each grid cell attributes: centroid_x, centroid_y, cluster_id, sum, n_cells, area
        """
        clusters_for_column = self.clustering.by_column[cluster_column]
        df = _gpd_GeoDataFrame({
            'centroid_x': [cluster.centroid[0] for cluster in clusters_for_column.clusters],
            'centroid_y': [cluster.centroid[1] for cluster in clusters_for_column.clusters],
            'cluster_id': [cluster.id for cluster in clusters_for_column.clusters],
            'sum': [cluster.total for cluster in clusters_for_column.clusters],
            "n_cells": [cluster.n_cells for cluster in clusters_for_column.clusters],
            'area': [cluster.area for cluster in clusters_for_column.clusters],
            },
            geometry = [cluster.geometry for cluster in clusters_for_column.clusters],
            crs=self.proj_crs)
        
        target_crs = self.input_crs if target_crs=='initial' else self.proj_crs if target_crs=='local' else target_crs
        if target_crs != self.proj_crs:
            transformer = Transformer.from_crs(crs_from=self.proj_crs, crs_to=target_crs, always_xy=True)
            df['centroid_x'], df['centroid_y'] = transformer.transform(df['centroid_x'], df['centroid_y'])
            df.to_crs(target_crs, inplace=True)
        
        return df
    #

    def save_full_grid(
            self,
            filename:str="full_grid",
            file_format:str=['shp','csv'][0],
            target_crs:str=['initial','local','EPSG:4326'][0],
    ):
        """save geopandas.DataFrame with entry for each grid cell with attributes on their Polygon, centroid, sum of indicator(s), and cluster id
        
        filename (str):
            name of the output file excluding file format extension. It can contain full path like 'output_folder/fname' (default='full_grid')
        file_format (str):
            format in which the file shall be saved. Currently available options are 'shp' and 'csv'. Extension will be appended to filename. (default='shp')
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        
        Returns:
        df (geopandas.GeoDataFrame):
            with entry for each grid cell
        """
        
        df = self.create_full_grid_df(target_crs=target_crs, max_column_name_length=10 if file_format=='shp' else 20)
        # save
        filename = filename +'.'+file_format
        if file_format == 'shp':
            df.to_file(filename, driver="ESRI Shapefile", index=False)
        elif file_format == 'csv':
            df.to_csv(filename, index=False)
        else:
            raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        return df
    #

    def save_sparse_grid(
            self,
            filename:str="sparse_grid",
            file_format:str=['shp','csv'][0],
            target_crs:str=['initial','local','EPSG:4326'][0],
        ):
        """save geopandas.GeoDataFrame with entry for grid cells that either has points inside or is part of a cluster with attributes on their Polygon, centroid, sum of indicator(s), and cluster id
        
        Args:
        filename (str):
            name of the output file excluding file format extension. It can contain full path like 'output_folder/fname' (default='sparse_grid')
        file_format (str):
            format in which the file shall be saved. Currently available options are 'shp' and 'csv'. Extension will be appended to filename. (default='shp')
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        
        Returns:
        df (geopandas.GeoDataFrame):
            with entry for each grid cell that is non empty or part of cluster
        """
        
        df = self.create_sparse_grid_df(
            target_crs=target_crs, max_column_name_length=10 if file_format=='shp' else 20
        )
        # save
        filename = filename +'.'+file_format
        if file_format == 'shp':
            df.to_file(filename, driver="ESRI Shapefile", index=False)
        elif file_format == 'csv':
            df.to_csv(filename, index=False)
        else:
            raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        return df
    #

    def save_cell_clusters_for_column(
            self,
            cluster_column:str,
            filename:str="grid_clusters",
            file_format:str=['shp','csv'][0],
            target_crs:str=['initial','local','EPSG:4326'][0],
    ):
        """Save geopandas.GeoDataFrame that has one entry for each clusters including attributes on its polygon geometry, column totals, n_cells, id  

        filename (str):
            name of the output file excluding file format extension. It can contain full path like 'output_folder/fname' (default='grid_clusters')
        file_format (str):
            format in which the file shall be saved. Currently available options are 'shp' and 'csv'. Extension will be appended to filename. (default='shp')
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        
        Returns:
        df (geopandas.GeoDataFrame)
            with one entry for each cluster
        """
        df = self.create_clusters_df_for_column(cluster_column=cluster_column, target_crs=target_crs)
        filename = filename+'.'+file_format
        if file_format == 'shp':
            df.to_file(filename, driver="ESRI Shapefile", index=False)
        elif file_format == 'csv':
            df.to_csv(filename, index=False)
        else:
            raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        return df
        #
    #
    def save_cell_clusters(
            self,
            filename:str="grid_clusters",
            file_format:str=['shp','csv'][0],
            target_crs:str=['initial','local','EPSG:4326'][0],
        ):
        """For each cluster column saves a geopandas.GeoDataFrame that has one entry for each clusters including attributes on its polygon geometry, column totals, n_cells, id  

        filename (str or list):
            name of the output file excluding file format extension. If there are more than 1 cluster column it will append the column name to the file. You can also provide a list of filenames to contain the filename indvidually. It can contain full path like 'output_folder/fname' (default='grid_clusters')
        file_format (str):
            format in which the file shall be saved. Currently available options are 'shp' and 'csv'. Extension will be appended to filename. (default='shp')
        target_crs (str):
            crs in which data shall be projected. If 'initial' then it will be projected in same crs as input data. If 'local' a local projection will be used. Otherwise specify the target crs directly like 'EPSG:4326' (default='initial') 
        
        Returns:
        dfs (_DFList)
            List of GeoDataFrames, one per cluster column. Also supports dict-style access by
            cluster column name: ``dfs['employment_cluster_sum_750']``, ``dfs.keys()``, ``dfs.items()``.
            Existing code using ``dfs[0]`` or ``for df in dfs`` is unchanged.
        """
        dfs = _DFList()
        filenames = filename if type(filename) == list else [filename + (("_"+cluster_column) if len(self.clustering.by_column) > 1 else '') for cluster_column in self.clustering.by_column]
        for (cluster_column, clusters), filename in zip(self.clustering.by_column.items(), filenames):
            df = self.save_cell_clusters_for_column(cluster_column=cluster_column, filename=filename, file_format=file_format, target_crs=target_crs)
            dfs._append(cluster_column, df)
        return dfs
    #
    #


#

class ExcludedArea:
    def __init__(self,excluded_area_geometry_or_list, grid:Grid):
        # recursively split exluded area geometry along grid 
        # then sort it into grid cell
        
        pass
#





"""
In-process progress bars for build_disk_region_lookups and search_and_aggregate.

Stage weights are placeholder fractions of total time that sum to 1.0. They will be
calibrated against real timing once benchmark data is collected. Within a session,
total-time estimates self-correct after each completed run using an exponential
moving average (new = 0.75 * old + 0.25 * actual), so runs on a faster or slower
machine converge toward reality within 2-3 calls.

To disable globally::

    import aabpl.utils.progress as aabpl_progress
    aabpl_progress.SHOW_PROGRESS = False
"""
import sys as _sys
import time as _time
import contextvars as _contextvars
try:
    from aabpl import config as _cfg
except Exception:
    _cfg = None

# ---------------------------------------------------------------------------
# Outer-call tracking — used to suppress sub-progress bars when a combined
# top-level progress bar (RadiusSearchProgress / DetectClusterProgress) is
# already active.  ContextVar is thread- and asyncio-task-safe.
# ---------------------------------------------------------------------------
_OUTER_PROGRESS: _contextvars.ContextVar = _contextvars.ContextVar(
    '_aabpl_outer_progress', default=None
)

# Module-level reference to whichever bar is currently rendering a \r line.
# Any bar sets this in start() and clears it in done()/cancel().
# progress_print() uses it to erase the line before printing and redraw after.
_ACTIVE_BAR = None

# Wall time: used for ETA display (what the user actually waits for).
_wall = _time.perf_counter
# CPU time: used for EMA learning (immune to parallel load on a busy machine).
_cpu = _time.process_time

# Set to False to suppress all progress bars regardless of the silent parameter.
SHOW_PROGRESS: bool = True

# Progress bars are suppressed for runs estimated to finish under this many seconds.
_MIN_SECONDS_TO_SHOW: float = 5.0

# Weight given to the new observation vs the old estimate (0.25 → 75% old, 25% new).
_EMA_ALPHA: float = 0.25

_BAR_WIDTH = 18
_LINE_WIDTH = 115  # must exceed the longest line any renderer can produce


_RATIO_EMA_ALPHA: float = 0.25


def progress_print(*args, **kwargs) -> None:
    """
    Print a message without disrupting the active progress bar.

    Clears the current bar line, prints the message (with a newline so it
    stays visible), then redraws the bar on the line below.  Falls back to
    plain ``print`` when no bar is active.

    Works for all bar types (DiskRegionProgress, SearchProgress,
    RadiusSearchProgress, SweepProgress) — whichever called start() most
    recently is tracked in _ACTIVE_BAR.
    """
    bar = _ACTIVE_BAR
    if bar is not None and getattr(bar, '_active', False):
        _sys.stdout.write(f"\r{' ' * _LINE_WIDTH}\r")
        _sys.stdout.flush()
        print(*args, **kwargs)
        bar._render()
    else:
        print(*args, **kwargs)


def _update_ratio(wall: float, cpu: float) -> None:
    """Update WALL_TO_CPU_RATIO EMA from a (wall_elapsed, cpu_elapsed) observation."""
    try:
        if _cfg is not None and cpu > 0.1:
            obs = wall / cpu
            _cfg.WALL_TO_CPU_RATIO = (1 - _RATIO_EMA_ALPHA) * _cfg.WALL_TO_CPU_RATIO + _RATIO_EMA_ALPHA * obs
    except Exception:
        pass


def _ratio() -> float:
    return _cfg.WALL_TO_CPU_RATIO


def _fmt_eta(seconds: float) -> str:
    """Format ETA as '~45s left', '~2m07s left', or '~1h23m left'."""
    s = max(0, int(round(seconds)))
    if s < 60:
        return f"~{s}s left"
    m, s = divmod(s, 60)
    if m < 60:
        return f"~{m}m{s:02d}s left"
    h, m = divmod(m, 60)
    return f"~{h}h{m:02d}m left"


# ---------------------------------------------------------------------------
# build_disk_region_lookups progress
# ---------------------------------------------------------------------------

_STAGE_LABELS = [
    "Classify cells",
    "Init triangle + boundary checks",
    "Split regions",
    "Finalize region cells",
    "Expand to 8 sectors",
    "Build region raster",
    "Expand raster + check structures",
    "Build lookup tables",
]
_STAGE_WEIGHTS = [0.02, 0.04, 0.45, 0.20, 0.05, 0.07, 0.10, 0.07]
assert abs(sum(_STAGE_WEIGHTS) - 1.0) < 1e-9, "Stage weights must sum to 1.0"

# Placeholder total-time estimate for build_disk_region_lookups (seconds).
# Updated within the session by EMA after each completed run.
_BUILD_EST_SECONDS: float = 180.0


class DiskRegionProgress:
    """
    Progress bar for a single build_disk_region_lookups call.

    Suppressed automatically when:
    - ``silent=True``, or
    - module-level ``SHOW_PROGRESS`` is ``False``, or
    - the current session estimate is below 5 s.

    Everything is written with ``\\r`` (no ``\\n``) so the line is fully erased
    on completion or cancellation — no lingering output.
    """

    def __init__(self, silent: bool, r_over_spacing: float, nest_depth: int):
        self._r_over_s = r_over_spacing
        self._nest_depth = nest_depth
        self._stage: int = -1
        self._desc: str = ""
        self._active: bool = False
        self._base_silent = silent

    def start(self) -> None:
        global _ACTIVE_BAR
        if self._base_silent or not SHOW_PROGRESS or _BUILD_EST_SECONDS < _MIN_SECONDS_TO_SHOW or _OUTER_PROGRESS.get() is not None:
            self._active = False
            return
        self._active = True
        _ACTIVE_BAR = self
        self._run_start_wall = _wall()
        self._run_start_cpu = _cpu()
        self._stage = -1
        self._render()

    def step(self, desc: str) -> None:
        if not self._active:
            return
        self._stage += 1
        self._desc = desc
        self._render()

    def done(self) -> None:
        global _BUILD_EST_SECONDS, _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        _sys.stdout.write("\r" + " " * _LINE_WIDTH + "\r")
        _sys.stdout.flush()
        wall_actual = _wall() - self._run_start_wall
        cpu_actual  = _cpu()  - self._run_start_cpu
        _BUILD_EST_SECONDS = (1 - _EMA_ALPHA) * _BUILD_EST_SECONDS + _EMA_ALPHA * cpu_actual
        _update_ratio(wall_actual, cpu_actual)

    def cancel(self) -> None:
        """Erase the progress line without updating the EMA (use on error)."""
        global _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        _sys.stdout.write("\r" + " " * _LINE_WIDTH + "\r")
        _sys.stdout.flush()

    def _render(self) -> None:
        try:
            elapsed = _wall() - self._run_start_wall
            p = sum(_STAGE_WEIGHTS[: self._stage + 1]) if self._stage >= 0 else 0.0
            eta_elapsed = max(0.0, elapsed / p - elapsed) if p > 0 and elapsed > 0 else _BUILD_EST_SECONDS * _ratio()
            eta_estimate = _BUILD_EST_SECONDS * _ratio() * (1.0 - p)
            blend = min(1.0, p / 0.3)
            eta = (1.0 - blend) * eta_estimate + blend * eta_elapsed
            pct = int(p * 100)
            filled = int(_BAR_WIDTH * p)
            bar = "=" * filled + ">" + " " * (_BAR_WIDTH - filled - 1)
            desc = (self._desc[:16] + "..") if len(self._desc) > 18 else self._desc
            remaining = _fmt_eta(eta)
            info = f"r/s={round(self._r_over_s, 2)} nd={self._nest_depth}"
            line = f"  [{bar}] {pct:3d}%  {info:<14}  {desc:<18}  {remaining:<14}"
            _sys.stdout.write(f"\r{line:<{_LINE_WIDTH}}")
            _sys.stdout.flush()
        except Exception:
            self._active = False


# ---------------------------------------------------------------------------
# search_and_aggregate (disk search loop) progress
# ---------------------------------------------------------------------------

# Placeholder total-time estimate for search_and_aggregate (seconds).
_SEARCH_EST_SECONDS: float = 180.0


class SearchProgress:
    """
    Progress bar for a single search_and_aggregate call.

    Designed for minimal overhead inside a tight per-point loop.  The caller
    drives updates using ``next_threshold``:

        prog = SearchProgress(silent=silent, n_pts=n_pts)
        prog.start()
        _thresh = prog.next_threshold   # copy to local for fast comparison
        for i, ... in enumerate(...):
            ...
            if i >= _thresh:
                _thresh = prog.update(i)
        prog.done()

    ``update`` is called at most 10 times per run, so its stdout cost is
    negligible.  The per-iteration overhead is a single integer comparison.
    When suppressed ``next_threshold`` is ``n_pts + 1`` so the condition is
    never true and there is zero overhead inside the loop.

    Everything is written with ``\\r`` (no ``\\n``) so the line is fully erased
    on completion — no lingering output.
    """

    def __init__(self, silent: bool, n_pts: int):
        self._n_pts = n_pts
        self._base_silent = silent
        self._active: bool = False
        # default: threshold beyond any valid i so the check never fires
        self.next_threshold: int = n_pts + 1

    # Target wall-time interval between redraws (seconds).
    _UPDATE_INTERVAL: float = 2.0

    def start(self) -> None:
        global _ACTIVE_BAR
        if self._base_silent or not SHOW_PROGRESS or _OUTER_PROGRESS.get() is not None:
            self._active = False
            return
        self._run_start_wall = _wall()
        self._run_start_cpu = _cpu()
        self._last_update_wall = self._run_start_wall
        self._last_update_cpu  = self._run_start_cpu
        self._last_update_i    = 0
        # Initial threshold: whichever fires first — 10 % of points or estimated
        # time for one update interval (so first redraw is never later than that).
        est_wall = _SEARCH_EST_SECONDS * _ratio()
        pts_per_interval = max(1, int(self._n_pts * self._UPDATE_INTERVAL / max(est_wall, 1e-9)))
        self.next_threshold = min(max(1, self._n_pts // 10), pts_per_interval)
        if est_wall >= _MIN_SECONDS_TO_SHOW:
            self._active = True
            _ACTIVE_BAR = self
            remaining = _fmt_eta(est_wall)
            bar = ">" + " " * (_BAR_WIDTH - 1)
            n_str = f"{self._n_pts:,} pts"
            line = f"  [{bar}]   0%  {n_str:<14}                    {remaining:<14}"
            _sys.stdout.write(f"\r{line:<{_LINE_WIDTH}}")
            _sys.stdout.flush()
        else:
            self._active = False  # lazily activated in update() if run is slow

    def update(self, i: int) -> int:
        """Redraw bar and return the next threshold value."""
        if not hasattr(self, '_run_start_wall'):
            return self._n_pts + 1
        now     = _wall()
        now_cpu = _cpu()
        elapsed = now - self._run_start_wall
        if not self._active:
            if elapsed < _MIN_SECONDS_TO_SHOW:
                # Keep checking every interval worth of points
                step = max(1, self.next_threshold - self._last_update_i)
                self._last_update_i = i
                return i + step
            global _ACTIVE_BAR
            self._active = True  # run is slower than estimated — show bar now
            _ACTIVE_BAR = self
        try:
            p = i / self._n_pts
            est_wall = _SEARCH_EST_SECONDS * _ratio()
            eta_elapsed = max(0.0, elapsed / p - elapsed) if p > 0 and elapsed > 0 else est_wall
            eta_estimate = est_wall * (1.0 - p)
            blend = min(1.0, p / 0.3)
            eta = (1.0 - blend) * eta_estimate + blend * eta_elapsed
            pct = int(p * 100)
            filled = int(_BAR_WIDTH * p)
            bar = "=" * filled + ">" + " " * (_BAR_WIDTH - filled - 1)
            remaining = _fmt_eta(eta)
            n_str = f"{self._n_pts:,} pts"
            line = f"  [{bar}] {pct:3d}%  {n_str:<14}                    {remaining:<14}"
            _sys.stdout.write(f"\r{line:<{_LINE_WIDTH}}")
            _sys.stdout.flush()
        except Exception:
            self._active = False
            return self._n_pts + 1
        # Update wall/cpu ratio from this interval, then recalibrate threshold
        dt     = now - self._last_update_wall
        dt_cpu = now_cpu - self._last_update_cpu
        di     = max(1, i - self._last_update_i)
        _update_ratio(dt, dt_cpu)
        self._last_update_wall = now
        self._last_update_cpu  = now_cpu
        self._last_update_i    = i
        pts_per_interval = max(1, int(di / max(dt, 1e-9) * self._UPDATE_INTERVAL))
        return i + pts_per_interval

    def done(self) -> None:
        global _SEARCH_EST_SECONDS, _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        _sys.stdout.write("\r" + " " * _LINE_WIDTH + "\r")
        _sys.stdout.flush()
        wall_actual = _wall() - self._run_start_wall
        cpu_actual  = _cpu()  - self._run_start_cpu
        _SEARCH_EST_SECONDS = (1 - _EMA_ALPHA) * _SEARCH_EST_SECONDS + _EMA_ALPHA * cpu_actual
        _update_ratio(wall_actual, cpu_actual)


# ---------------------------------------------------------------------------
# Combined top-level progress bars
# ---------------------------------------------------------------------------

class _CombinedProgress:
    """
    Single-line progress bar for a top-level user-facing call.

    Sub-progress bars (DiskRegionProgress, SearchProgress) suppress themselves
    when ``_OUTER_PROGRESS`` is set, so only this line is visible.

    Subclasses define:
      _STAGE_WEIGHTS  – ordered list of (stage_name, cumulative_fraction) pairs
      _est_seconds    – class-level float: current EMA estimate of total wall time.
                        Updated after each completed run so ETA self-corrects.

    Lifecycle::

        prog = RadiusSearchProgress(silent=silent, n_pts=len(pts))
        token = _OUTER_PROGRESS.set(prog)
        prog.start()
        try:
            prog.step("assigning target")
            ...
            prog.step("searching")
            ...
        finally:
            _OUTER_PROGRESS.reset(token)
            prog.done()
    """

    _label: str = "aabpl"
    _STAGE_WEIGHTS: list = []
    _est_seconds: float = 60.0  # placeholder; updated by EMA after each run

    def __init__(self, silent: bool, n_pts: int, n_tgt: int = None):
        self._n_pts = n_pts
        self._n_tgt = n_tgt
        self._silent = silent
        self._active = False
        self._stage = ""
        self._progress = 0.0
        self._stage_map = {name: frac for name, frac in self._STAGE_WEIGHTS}

    def start(self) -> None:
        global _ACTIVE_BAR
        if self._silent or not SHOW_PROGRESS:
            return
        self._active = True
        _ACTIVE_BAR = self
        self._t0 = _wall()
        self._cpu0 = _cpu()
        self._progress = 0.0
        self._render()

    def step(self, stage: str) -> None:
        if not self._active:
            return
        self._stage = stage
        self._progress = self._stage_map.get(stage, self._progress)
        self._render()

    def done(self) -> None:
        global _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        wall_elapsed = _wall() - self._t0
        cpu_elapsed  = _cpu()  - self._cpu0
        type(self)._est_seconds = (1 - _EMA_ALPHA) * type(self)._est_seconds + _EMA_ALPHA * wall_elapsed
        _update_ratio(wall_elapsed, cpu_elapsed)
        _sys.stdout.write("\r" + " " * _LINE_WIDTH + "\r")
        _sys.stdout.flush()

    def cancel(self) -> None:
        global _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        _sys.stdout.write("\r" + " " * _LINE_WIDTH + "\r")
        _sys.stdout.flush()

    def _render(self) -> None:
        try:
            elapsed = _wall() - self._t0
            p = self._progress
            est = type(self)._est_seconds
            eta_elapsed = max(0.0, elapsed / p - elapsed) if p > 0 and elapsed > 0 else est
            eta_estimate = est * (1.0 - p)
            blend = min(1.0, p / 0.3)
            eta = (1.0 - blend) * eta_estimate + blend * eta_elapsed
            filled = int(_BAR_WIDTH * p)
            bar = "=" * filled + (">" if filled < _BAR_WIDTH else "") + " " * max(0, _BAR_WIDTH - filled - 1)
            pct = int(p * 100)
            if self._n_tgt is not None and self._n_tgt != self._n_pts:
                n_str = f"{self._n_pts:,}->{self._n_tgt:,}"
            else:
                n_str = f"{self._n_pts:,} pts"
            stage = (self._stage[:18] + "..") if len(self._stage) > 20 else self._stage
            remaining = _fmt_eta(eta)
            line = f"  {self._label}  [{bar}] {pct:3d}%  {stage:<22}  {n_str:<14}  {elapsed:.1f}s  {remaining:<12}"
            _sys.stdout.write(f"\r{line:<{_LINE_WIDTH}}")
            _sys.stdout.flush()
        except Exception:
            self._active = False


class RadiusSearchProgress(_CombinedProgress):
    """Combined progress bar for a top-level ``radius_search`` call."""
    _label = "radius_search  "
    _est_seconds: float = 30.0  # placeholder: ~30s for a medium dataset
    _STAGE_WEIGHTS = [
        ("initializing",     0.00),
        ("assigning target", 0.05),
        ("assigning source", 0.10),
        ("searching",        0.15),
    ]


class DetectClusterProgress(_CombinedProgress):
    """Combined progress bar for a top-level ``detect_cluster_pts`` call."""
    _label = "detect_clusters"
    _est_seconds: float = 90.0  # placeholder: ~90s (two searches + null distribution)
    _STAGE_WEIGHTS = [
        ("initializing",      0.00),
        ("assigning target",  0.03),
        ("null distribution", 0.08),
        ("assigning source",  0.25),
        ("searching",         0.30),
        ("labeling clusters", 0.90),
    ]


# ---------------------------------------------------------------------------
# Sweep progress bar (benchmarking)
# ---------------------------------------------------------------------------

class SweepProgress:
    """
    Single-line loading bar for run_sweep iterations.

    Usage::

        prog = SweepProgress(n_total=120)
        prog.start()
        for ...:
            prog.update(done, label="r=500 s=100 nd=2 tn=1")
            ...
        prog.done()
    """

    def __init__(self, n_total: int):
        self._n = n_total
        self._active = False

    def start(self) -> None:
        global _ACTIVE_BAR
        self._active = True
        _ACTIVE_BAR = self
        self._t0 = _wall()
        self._done = 0
        self._label = ""
        self._render()

    def update(self, done: int, label: str = "") -> None:
        if not self._active:
            return
        self._done = done
        self._label = label
        self._render()

    def clear(self) -> None:
        """Erase the bar line so external prints appear on a clean line."""
        if not self._active:
            return
        _sys.stdout.write(f"\r{' ' * _LINE_WIDTH}\r")
        _sys.stdout.flush()

    def redraw(self) -> None:
        """Redraw the bar after external prints have moved the cursor."""
        if not self._active:
            return
        self._render()

    def done(self) -> None:
        global _ACTIVE_BAR
        if not self._active:
            return
        self._active = False
        if _ACTIVE_BAR is self:
            _ACTIVE_BAR = None
        _sys.stdout.write(f"\r{' ' * _LINE_WIDTH}\r")
        elapsed = _wall() - self._t0
        _sys.stdout.write(f"  sweep done  {self._n} runs  {elapsed:.1f}s\n")
        _sys.stdout.flush()

    def _render(self) -> None:
        elapsed = _wall() - self._t0
        p = self._done / self._n if self._n > 0 else 0.0
        eta = (elapsed / p - elapsed) if p > 0 and elapsed > 0 else 0.0
        pct = int(p * 100)
        filled = int(_BAR_WIDTH * p)
        bar = "=" * filled + (">" if filled < _BAR_WIDTH else "") + " " * max(0, _BAR_WIDTH - filled - 1)
        label_str = (self._label[:22] + "..") if len(self._label) > 24 else self._label
        eta_str = _fmt_eta(eta) if p > 0 else ""
        line = f"  [{bar}] {pct:3d}%  {self._done}/{self._n}  {label_str:<26}  {elapsed:.1f}s  {eta_str}"
        _sys.stdout.write(f"\r{line:<{_LINE_WIDTH}}")
        _sys.stdout.flush()



from numpy import (
    array as _np_array, 
    unique as _np_unique, 
    linspace, invert, flip, transpose, 
    concatenate, 
    sign as _np_sign, 
    zeros, min, max, equal, where, 
    logical_or, logical_and, all, newaxis
)
from pandas import DataFrame as _pd_DataFrame
from matplotlib.pyplot import (subplots as _plt_subplots, figure as _plt_figure)
from matplotlib.patches import Circle as _plt_Circle, Rectangle as _plt_Rectangle, Polygon as _plt_Polygon
from matplotlib.figure import Figure as _plt_Figure
from matplotlib.axes._axes import Axes as _plt_Axes
from aabpl.utils.misc import ( flatten_list, )


def illustrate_point_disk(
    grid:dict,
    shared_cntd_cells:list,
    shared_ovlpd_cells:list,
    distinct_cntd_cells:list,
    distinct_ovlpd_cells:list,
    pts_xy_in_cells_ovlpd_by_pt_region:_np_array,
    pts_xy_in_radius:_np_array,
    pts_xy_in_cell_cntd_by_pt_region:list,
    pts_source:_pd_DataFrame,
    pts_target:_pd_DataFrame,
    region_id,
    home_cell:tuple,
    r:float=750,
    sum_names:list=['employment'],
    y:str='proj_lat',
    x:str='proj_lon',
    **plot_kwargs,
):

    """
    Illustrate method
    """
    # specify default plot kwargs and add defaults
    plot_kwargs = {
        'fig':None,
        'ax':None,
        's':0.8,
        'color':'#eaa',
        'figsize': (20,30),
        **plot_kwargs
    }
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    ax = plot_kwargs.pop('ax')
    pt_id = plot_kwargs.pop('pt_id')
    sums_within_disk = plot_kwargs.pop('sums_within_disk', None)
    display_cell_region_id = plot_kwargs.pop('cell_region_id', region_id)
    # optional output path (accepts savevig / savefig / filename); popped so it
    # does not leak into the scatter/patch kwargs below.
    save_path = (plot_kwargs.pop('savevig', None) or plot_kwargs.pop('savefig', None)
                 or plot_kwargs.pop('filename', None))
    plot_kwargs.pop('nest_depth', None)
    # print("row:",pts_source.loc[pt_id])
    pt_x, pt_y = pts_source.loc[pt_id,[x,y]]
    # home_cell = grid.pt_id_to_row_col[pt_id]
    home_cell_centroid = grid.get_cell_centroid(*home_cell)
    hc_x,hc_y = home_cell_centroid
    ###### initialize plot  ######################
    colors = ["#1625ff","#12b400","#ff8c00", '#ff0000']
    if fig is None:
        fig, axs = _plt_subplots(1,1, figsize=figsize)
    elif type(fig) != _plt_Figure:
        raise TypeError
    # [(cells_ovlpd_by_pt_region, pt_in_radius) for pt_maybe_in_radius, pt_in_radius in zip(cells_ovlpd_by_pt_region, pts_in_radius)]
    # grid.search.target.
    ################################################################################################################
    ax = axs#[0]
    # print("(pt_x, pt_y)",(pt_x, pt_y))
    # print("cells", [ ((c-.5)*grid._search_spacing+grid.total_bounds.xmin, (row-.5)*grid._search_spacing+grid.total_bounds.ymin) for row,c in cells_cntd_by_pt_cell])
    # print(
    #     [[(grid.get_cell_centroid(row,col), color) for lvl,(row,col) in cells] for cells,color in zip(
    #     [cells_cntd_by_pt_cell, cells_cntd_by_pt_region, cells_ovlpd_by_pt_region],
    #     ['blue','green', 'red'])])
    # print(flatten_list(
    #     [[(grid.get_cell_centroid(row,col), color) for lvl,(row,col) in cells] for cells,color in zip(
    #     [cells_cntd_by_pt_cell, cells_cntd_by_pt_region, cells_ovlpd_by_pt_region],
    #     ['blue','green', 'red'])]))
    # print("distinct_ovlpd_cells",distinct_ovlpd_cells)
    # len(shared_cntd_cells), len(distinct_cntd_cells), etc.
    def _cell_centroid(lvl, row, col):
        # All levels: centroid = xmin + (col + 0.5) * spacing.
        # lvl=0: col is an integer grid index, so +0.5 gives the cell centre.
        # lvl>0: the anchor in check_nested_subcells is set as (index - 0.5), so
        #        the stored fractional coord is also 0.5 below the true centre;
        #        adding 0.5 corrects this for all nesting levels uniformly.
        return (
            grid.total_bounds.xmin + (col + 0.5) * grid._search_spacing,
            grid.total_bounds.ymin + (row + 0.5) * grid._search_spacing,
        )

    for (lvl,(cntrd_x,cntrd_y)), color, hatch in flatten_list(
        [[((lvl, _cell_centroid(lvl, row, col)), color, hatch) for lvl,(row,col) in cells] for cells,color,hatch in zip(
        [shared_cntd_cells, distinct_cntd_cells, shared_ovlpd_cells,distinct_ovlpd_cells],
        colors, range(4))]):
        # print("cntrd",cntrd)
        # print("grid._search_spacing",grid._search_spacing)
        # print("cntrd -( .5) * grid._search_spacing",(cntrd[0] -( .5) * grid._search_spacing, cntrd[1] -( .5) * grid._search_spacing))
        xy = (cntrd_x-2**(-lvl-1)*grid._search_spacing, cntrd_y-2**(-lvl-1)*grid._search_spacing)
        # print("+xy",xy)
        # hatches = ['*', '\\', '-', '/', '+', 'x', 'o', 'O', '.', '*']
        hatches = ['', '\\', '/', 'o', '.', 'x', '*']
        
        
        ax.add_patch(_plt_Rectangle(
            xy = xy, 
            hatch=hatches[(hatch*0+1*lvl)%len(hatches)],
            width=2**(-lvl)*grid._search_spacing, height=2**(-lvl)*grid._search_spacing, 
            linewidth=.7, facecolor=color, edgecolor=color, alpha=0.3
        ))
        ax.add_patch(_plt_Rectangle(
            xy = xy, 
            width=2**(-lvl)*grid._search_spacing, height=2**(-lvl)*grid._search_spacing, 
            linewidth=.7, facecolor='None', edgecolor=color, alpha=0.8
        ))
        # if lvl==1:
        #     ax.annotate(text=str((
        #     float(round((cntrd_x-grid.total_bounds.xmin)/grid._search_spacing-home_cell[1],5)),
        #     float(round((cntrd_y-grid.total_bounds.ymin)/grid._search_spacing-home_cell[0],5)),
        #     )), xy=xy, 
        #         horizontalalignment='center',
        #         backgroundcolor="#ffffff88",)
    cntrd_color = flatten_list(
        [[(_cell_centroid(lvl, row, col), color, lvl) for lvl,(row,col) in cells] for cells,color in zip(
        [shared_cntd_cells, distinct_cntd_cells, shared_ovlpd_cells,distinct_ovlpd_cells],
        colors)])
    ax.scatter(
        x=[cntrd[0] for cntrd,color, lvl in cntrd_color],
        y =[cntrd[1] for cntrd,color, lvl in cntrd_color],
        s=[fig.get_figheight()*250*(2**-lvl) for cntrd,color, lvl in cntrd_color], 
        linewidths=[2*(2**-lvl) for cntrd,color, lvl in cntrd_color],
        c=[color for cntrd,color,lvl in cntrd_color], marker='+', alpha=0.1)
    # ax.scatter(
    #     x=[cntrd[0] for cntrd,color, lvl in cntrd_color],
    #     y =[cntrd[1] for cntrd,color, lvl in cntrd_color],
    #     s=fig.get_figheight()*500, c=[color for cntrd,color,lvl in cntrd_color], marker='+', alpha=0.1)
    # flat_list = flatten_list(
    #     [[(grid.get_cell_centroid(row,col), color) for lvl,(row,col) in cells] for cells,color in zip(
    #     [nested_cells_cntd_by_pt_region, nested_cells_ovlpd_by_pt_region],
    #     ["#1d802a", "#a51414"])])
    # print("flat_list",flat_list)

    # add_grid_cell_rectangles_by_color(
    #     [cells_cntd_by_pt_cell, cells_cntd_by_pt_region, cells_ovlpd_by_pt_region],
    #     ['blue','green', 'red'],
    #     ax=ax, grid_spacing=grid._search_spacing,
    #     x_off=grid.total_bounds.xmin+grid._search_spacing/2,
    #     y_off=grid.total_bounds.ymin+grid._search_spacing/2,
    # )
    # Offset-region overlay (the family of disk centres for the point's region).
    # region_id is keyed differently from cell_region; draw it only when available.
    offset_region = grid.id_to_offset_regions.get(region_id) if region_id is not None else None
    if offset_region is not None:
        region_coords = offset_region.get_plot_coords()
        for region_x, region_y in region_coords:
            ax.add_patch(
                _plt_Circle(
                    xy=(region_x*grid._search_spacing+hc_x, region_y*grid._search_spacing + hc_y), radius=r,
                    facecolor="#000000"+(str(int(60/len(region_coords))) if int(60/len(region_coords))>=10 else '0'+str(int(60/len(region_coords)))),
                    edgecolor='#0006', linewidth=0.25))
        ax.add_patch(_plt_Polygon(
            [(region_x*grid._search_spacing+hc_x, region_y*grid._search_spacing + hc_y) for region_x, region_y in region_coords],
            facecolor="#000000",))
    ax.add_patch(_plt_Circle(xy=(pt_x, pt_y), radius=r, facecolor="#0000ff16",edgecolor='#00f',linewidth=2,))
    ax.add_patch(_plt_Circle(xy=(pt_x, pt_y), radius=r/40, alpha=0.6))
    # ax.add_patch(create_buffered_square_patch(side_length=grid._search_spacing, r=r, x_off=hc_x, y_off=hc_y))
    # ax.add_patch(create_debuffered_square_patch(side_length=grid._search_spacing, r=r, linewidth=2, x_off=hc_x, y_off=hc_y ))
    
    # ax.add_patch(create_trgl1_patch(side_length=grid._search_spacing/2, linewidth=2, x_off=hc_x, y_off=hc_y ))
    # ax.add_patch(create_buffered_trgl1_patch(side_length=grid._search_spacing/2, linewidth=2, x_off=hc_x, y_off=hc_y ))
    # ax.add_patch(create_debuffered_trgl1_patch(side_length=grid._search_spacing/2, linewidth=2, x_off=hc_x, y_off=hc_y ))
    # print('+++++', [(cells,color) for cells,color in zip(
    #     [cells_cntd_by_pt_cell, cells_cntd_by_pt_region, cells_ovlpd_by_pt_region],
    #     ['blue','green', 'red'])])
    # print('+++++', [[(grid.get_cell_centroid(row,col), color) for (row,col) in cells] for cells,color in zip(
    #     [cells_cntd_by_pt_cell, cells_cntd_by_pt_region, cells_ovlpd_by_pt_region],
    #     ['blue','green', 'red'])])

    # all pts
    ax.scatter(
        x=pts_target[x],
        y =pts_target[y],
        # c=pts_target['sc_nr'],
        # s=fig.get_figheight()/.2, cmap='viridis', marker='x')
        s=fig.get_figheight()/1, color='#777', marker='x')
    # pts in cntd cells
    ax.scatter(
        x=pts_xy_in_cell_cntd_by_pt_region[:,0],
        y =pts_xy_in_cell_cntd_by_pt_region[:,1],
        s=fig.get_figheight()/2, color='yellow', marker='o')
    # pts in ovlpd cells
    ax.scatter(
        x=pts_xy_in_cells_ovlpd_by_pt_region[:,0],
        y =pts_xy_in_cells_ovlpd_by_pt_region[:,1],
        s=fig.get_figheight()/2, color='red', marker='+')
    # pts in overlapped cells inside r
    ax.scatter(
        x=pts_xy_in_radius[:,0],
        y =pts_xy_in_radius[:,1],
        s=fig.get_figheight()/2, color='black', marker='o')
    
    # for (i, ax) in enumerate(axs):
    ax.set_xlim(pt_x-1.35*r,pt_x+1.35*r)
    ax.set_ylim(pt_y-1.35*r,pt_y+1.35*r)
    ax.set_aspect('equal', adjustable='box')

    display_lines = []
    if sums_within_disk is not None:
        for name, value in zip(sum_names, sums_within_disk):
            try:
                display_lines.append(f"{name}: {float(value):.6g}")
            except Exception:
                display_lines.append(f"{name}: {value}")
    else:
        display_lines = [str(name) for name in sum_names]

    display_lines.append(f"\ncell_region_id: {display_cell_region_id}")

    title_text = "  •  ".join(display_lines)
    ax.set_title(
        title_text,
        fontsize=40,
        pad=18,
        loc='left',
    )
    if save_path:
        if not str(save_path).lower().endswith(('.png', '.jpg', '.jpeg', '.pdf', '.svg')):
            save_path = str(save_path) + '.png'
        fig.savefig(save_path, dpi=150, bbox_inches='tight')
    #
#



from pandas import DataFrame as _pd_DataFrame
from math import floor as _math_floor
from pyproj import Transformer as _pyproj_Transformer
from shapely.ops import transform as _shapely_transform

def convert_wgs_to_utm(lon: float, lat: float):
    """Based on lat and lng, return best utm epsg-code"""
    # https://gis.stackexchange.com/a/269552
    # convert_wgs_to_utm function, see https://stackoverflow.com/a/40140326/4556479
    # see https://gis.stackexchange.com/a/127432/33092
    utm_band = str((_math_floor((lon + 180) / 6 ) % 60) + 1)
    if len(utm_band) == 1:
        utm_band = '0'+utm_band
    if lat >= 0:
        epsg_code = '326' + utm_band
        return epsg_code
    epsg_code = '327' + utm_band
    return epsg_code
#

def convert_MultiPolygon_crs(
        multipoly,
        initial_crs:str="EPSG:4326",
        target_crs:str="EPSG:4326",
):
    """Reprojects (Mulit-)Polygon from initial crs to target crs.

    """
    if initial_crs == target_crs:
        return multipoly
    project = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to=target_crs, always_xy=True).transform
    try: 
        multipoly_transformed = _shapely_transform(project, multipoly)
    except:
        from aabpl.utils.progress import progress_print
        progress_print("ERROR in reprojecting sample_area "+str(type(multipoly))+" from "+initial_crs+" to "+target_crs+". Ensure that both crs are valid for coordinates of the sample_area.")
        multipoly_transformed = _shapely_transform(project, multipoly)
    return multipoly_transformed

def convert_coords_to_local_crs(
        pts,
        x:str='lon',
        y:str='lat',
        proj_x:str='proj_lon',
        proj_y:str='proj_lat',
        initial_crs:str="EPSG:4326",
        target_crs:str='auto',
        silent:bool=False,
) -> str:
    """Reprojects coordinates into target crs. Modifies DataFrame and returns string of local_crs. If non specified it chooses best crs based on the mean coordinate.

    """
    tgt_was_auto = target_crs == 'auto'
    if tgt_was_auto:
        if initial_crs != "EPSG:4326":
            # Convert to WGS84 first to determine the best UTM zone, then pick local_crs
            transformer = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to="EPSG:4326", always_xy=True)
            x_wgs,y_wgs = transformer.transform(pts[x], pts[y])
            local_crs = 'EPSG:'+str(convert_wgs_to_utm(sum(x_wgs)/len(x_wgs), sum(y_wgs)/len(y_wgs)))
        else:
            local_crs = 'EPSG:'+str(convert_wgs_to_utm(*pts[[x,y]].mean(axis=0)))
    else:
        local_crs = target_crs
    transformer = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to=local_crs, always_xy=True)
    pts[proj_x],pts[proj_y] = transformer.transform(pts[x], pts[y])
    if initial_crs != local_crs and (silent==False or tgt_was_auto):
        from aabpl.utils.progress import progress_print
        progress_print("Reproject from " +str(initial_crs)+' to '+local_crs)
    return local_crs
#

def _find_matching_col(pts, values):
    """Return the name of the first column in pts whose float64 values exactly match `values`, or None."""
    import numpy as _np
    for col in pts.columns:
        try:
            if pts[col].values.dtype.kind in ('f', 'i', 'u') and _np.array_equal(pts[col].values.astype(float), values):
                return col
        except Exception:
            pass
    return None

def convert_pts_to_crs(
    pts:_pd_DataFrame=None,
    x:str='lon',
    y:str='lat',
    initial_crs:str='EPSG:4326',
    target_crs:str='auto',
    silent:bool=False,
):
    if target_crs is None:
        return x, y, initial_crs

    # Determine local_crs without writing to pts yet.
    tgt_was_auto = target_crs == 'auto'
    if tgt_was_auto:
        if initial_crs != "EPSG:4326":
            transformer_wgs = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to="EPSG:4326", always_xy=True)
            x_wgs, y_wgs = transformer_wgs.transform(pts[x], pts[y])
            local_crs = 'EPSG:' + str(convert_wgs_to_utm(sum(x_wgs)/len(x_wgs), sum(y_wgs)/len(y_wgs)))
        else:
            local_crs = 'EPSG:' + str(convert_wgs_to_utm(*pts[[x, y]].mean(axis=0)))
    else:
        local_crs = target_crs

    if local_crs == initial_crs:
        # Coords are already in the target CRS — x/y are usable as-is.
        return x, y, local_crs

    # Compute projected values so we can check whether they already exist in pts.
    from numpy import array as _np_array
    transformer = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to=local_crs, always_xy=True)
    x_proj, y_proj = transformer.transform(pts[x].values, pts[y].values)
    x_proj = _np_array(x_proj, dtype=float)
    y_proj = _np_array(y_proj, dtype=float)

    # Reuse an existing column if it already contains the exact projected values.
    existing_x = _find_matching_col(pts, x_proj)
    existing_y = _find_matching_col(pts, y_proj)

    if existing_x is None:
        proj_x = next(('proj_x' + str(i) for i in [''] + list(range(len(pts.columns))) if 'proj_x' + str(i) not in pts.columns))
        pts[proj_x] = x_proj
    else:
        proj_x = existing_x

    if existing_y is None:
        proj_y = next(('proj_y' + str(i) for i in [''] + list(range(len(pts.columns))) if 'proj_y' + str(i) not in pts.columns))
        pts[proj_y] = y_proj
    else:
        proj_y = existing_y

    if not silent or tgt_was_auto:
        from aabpl.utils.progress import progress_print
        progress_print("Reproject from " + str(initial_crs) + ' to ' + local_crs)

    return proj_x, proj_y, local_crs
#

def convert_bounds_to_local_crs(
        xmin:float,
        xmax:float,
        ymin:float,
        ymax:float,
        initial_crs:str="EPSG:4326",
        target_crs:str='auto',
        silent:bool=False,
) -> tuple:
    """Reprojects coordinates into target crs. Modifies DataFrame and returns string of local_crs. If non specified it chooses best crs based on the mean coordinate.
    
    """
    bounds_corners_x = [xmin,xmax,xmax,xmin]
    bounds_corners_y = [ymin,ymax,ymax,ymin]
    
    bounds_corners_x = []
    bounds_corners_y = []
    for x in [xmin, (xmin + xmax)/2, xmax]:
        for y in [ymin, (ymin + ymax)/2, ymax]:
            bounds_corners_x.append(x)
            bounds_corners_y.append(y)
    if target_crs == 'auto':
        if initial_crs != "EPSG:4326":
            transformer = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to="EPSG:4326", always_xy=True)
            x_wgs,y_wgs = transformer.transform(bounds_corners_x, bounds_corners_y)
            local_crs = 'EPSG:'+str(convert_wgs_to_utm(sum(x_wgs)/len(x_wgs), sum(y_wgs)/len(y_wgs)))
        else:
            local_crs = 'EPSG:'+str(convert_wgs_to_utm(
                sum(bounds_corners_x)/len(bounds_corners_x), 
                sum(bounds_corners_y)/len(bounds_corners_y))
                )
    else:
        local_crs = target_crs
    transformer = _pyproj_Transformer.from_crs(crs_from=initial_crs, crs_to=local_crs, always_xy=True)
    xs_local,ys_local = transformer.transform(bounds_corners_x, bounds_corners_y)
    return local_crs, (min(xs_local), max(xs_local),min(ys_local), max(ys_local))
#


from matplotlib.pyplot import (subplots as _plt_subplots)
from matplotlib.pyplot import get_cmap as _plt_get_cmap
from matplotlib.pyplot import close as _plt_close
from matplotlib.colors import LogNorm as _plt_LogNorm, Normalize as _plt_Normalize, LinearSegmentedColormap as _plt_LinearSegmentedColormap, ListedColormap as _plt_ListedColormap, BoundaryNorm as _plt_BoundaryNorm
from matplotlib.patches import Patch as _plt_Patch
from numpy import array as _np_array, zeros as _np_zeros
from aabpl.illustrations.plot_utils import truncate_colormap, set_map_frame, plot_polygon
from shapely.geometry import Polygon as _shapely_Polygon, MultiPoint as _shapely_MultiPoint

def handle_plot_kwargs(default_kwargs:dict={}, **kwargs):
    plot_kwargs = {
        **default_kwargs
    }
    sub_group_kwargs = ['fig','ax','hlines','vlines','figsize','set_aspect','set_xticks','set_yticks','set_xlim','set_ylim','suptitle']
    # groups of kwargs from which only the last one shall be applied
    kwargs_groups = (('cmap','color','c'),)
    kwarg_to_group = {}
    for grp in kwargs_groups:
        for k in grp:
            if k not in kwarg_to_group:
                kwarg_to_group[k] = [x for x in grp if x!=k]
            else:
                kwarg_to_group[k].extend([x for x in grp if x!=k])
            #
        #
    #
    # 'hlines':{'color':'red', 'linewidth':1},
    # 'vlines':{'color':'red', 'linewidth':1},
    for kwarg, v in kwargs.items():
        if kwarg in kwargs_groups:
            for k in kwargs_groups[kwarg]:
                plot_kwargs.pop(k,None)
        plot_kwargs[kwarg] = v
    used_sub_groups = [k for k in sub_group_kwargs if k in plot_kwargs]
    def pop_all_subgroups(used_sub_groups=used_sub_groups, plot_kwargs=plot_kwargs):
        for k in used_sub_groups:
            plot_kwargs.pop(k,None)
    # plot_kwargs['pop_all'] = pop_all_subgroups
    return plot_kwargs


def create_plots_for_vars(
        grid,
        colnames:_np_array,
        filename:str="",
        save_kwargs:dict={},
        plot_kwargs:dict={},
        show:bool=True,
        display_dpi:int=100,
        r:float=None,
):
    """
    Scatter plot of source points coloured by the value of one or more columns.

    Parameters
    ----------
    grid
        Grid object returned by ``radius_search`` / ``detect_cluster_pts``.
    colnames : np.ndarray
        Array of column names to plot. One subplot per entry.
    filename : str
        Save path. Empty string skips saving (default ``''``).
    save_kwargs : dict
        Forwarded to ``fig.savefig()``.
        Defaults: ``{'dpi': 300, 'bbox_inches': 'tight'}``.
    plot_kwargs : dict
        Supported keys:
        - ``figsize`` : tuple — figure size (default ``(10*n_cols, 8*n_rows)``)
        - ``cmap`` : str — colormap name (default ``'Reds'``)
        - ``s`` : float — scatter marker size (default auto)
        - ``vmin``, ``vmax`` : float — colormap range
        - ``sample_area_colors`` : list[str, str] — colours for [valid, excluded] area
        - ``sample_area_linewidth`` : float — border linewidth of the valid-area polygon (default ``0.2``)
        - ``fig``, ``axs`` — existing Figure/Axes to draw into
    show : bool
        Display the figure (default ``True``).
    display_dpi : int
        Resolution for inline/screen display (default ``100``).
        Use ``save_kwargs={'dpi': 300}`` to control saved-file resolution.

    Returns
    -------
    matplotlib.figure.Figure
    """
    nrows = colnames.shape[0]
    ncols = 1 if len(colnames.shape)==1 else colnames.shape[1]
    # Callers (e.g. grid.plot.vars / grid.plot.cluster_pts) forward **plot_kwargs;
    # extract control args that must not leak into ax.scatter().
    show = plot_kwargs.pop('show', show)
    display_dpi = plot_kwargs.pop('display_dpi', display_dpi)
    if 'save_kwargs' in plot_kwargs:
        save_kwargs = plot_kwargs.pop('save_kwargs')
    plot_kwargs.pop('filename', None)
    plot_kwargs.pop('colnames', None)
    # specify default plot kwargs and add defaults
    s = 1 if not 'figsize' in plot_kwargs else 0.1*plot_kwargs['figsize'][0]
    plot_kwargs = handle_plot_kwargs(default_kwargs={
        'fig': None,
        'axs': None,
        's': s,
        'cmap': 'Reds',
        'figsize': (10*ncols, 5*nrows),
        'additional_varnames':[],
        'sample_area_colors': ["#ffffff",  "#bedbe6"],
        'sample_area_linewidth': 0.2,
    }, **plot_kwargs)
    save_kwargs = {'dpi':300, 'bbox_inches':"tight", **save_kwargs}
    
    plot_kwargs.pop('color', None)
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    axs = plot_kwargs.pop('axs')
    
    sample_area_color, non_valid_area_color = plot_kwargs.pop('sample_area_colors')
    sample_area_lw = plot_kwargs.pop('sample_area_linewidth')
    
    additional_varnames = plot_kwargs.pop('additional_varnames')
    if len(additional_varnames)>nrows:
        # TODO this is not compelted
        nrows = len(additional_varnames)
    cmap_name = plot_kwargs.pop('cmap', 'Reds')
    cmap = truncate_colormap(_plt_get_cmap(cmap_name), 0.1, 1)
    if fig is None or axs is None:
        fig, axs = _plt_subplots(nrows, ncols, figsize=figsize, dpi=display_dpi,
                                  constrained_layout=True,
                                  sharex=(ncols == 1 and nrows > 1),
                                  sharey=(nrows == 1 and ncols > 1))
    if not grid.sample_grid_bounds is None and not grid.sample_area is None:
        non_valid_area = _shapely_Polygon([
            (grid.sample_grid_bounds[0], grid.sample_grid_bounds[1]),
            (grid.sample_grid_bounds[2], grid.sample_grid_bounds[1]),
            (grid.sample_grid_bounds[2], grid.sample_grid_bounds[3]),
            (grid.sample_grid_bounds[0], grid.sample_grid_bounds[3])
            ]).difference(grid.sample_area) 
        xmin, xmax, ymin, ymax = grid.total_bounds.xmin, grid.total_bounds.xmax, grid.total_bounds.ymin, grid.total_bounds.ymax,
    else:    
        xmin, xmax, ymin, ymax = grid.total_bounds.xmin, grid.total_bounds.xmax, grid.total_bounds.ymin, grid.total_bounds.ymax,
    _sc = getattr(grid, '_search_class', None) or grid.search
    x_col, y_col = _sc.source.x, _sc.source.y
    src = _sc.source
    xs = src.pts[x_col]
    ys = src.pts[y_col]
    for i, colname in enumerate(colnames.flat):
        # SELECT AX (IF MULTIPLE)
        ax = axs.flat[i] if nrows > 1 else axs
        row_idx = i // ncols
        col_idx = i % ncols

        # SET TITLE
        ax_title = (colname)
        ax.set_title(ax_title)
        # suppress redundant axis labels on shared axes
        if ncols == 1 and nrows > 1 and row_idx < nrows - 1:
            ax.set_xlabel('')
            ax.tick_params(labelbottom=False)
        if nrows == 1 and ncols > 1 and col_idx > 0:
            ax.set_ylabel('')
            ax.tick_params(labelleft=False)
        # CPOLOR NON SAMPLE AREA
        if not grid.sample_grid_bounds is None and not grid.sample_area is None:
            ax.set_facecolor(sample_area_color)
            cells_rndm_sample = grid.cells_rndm_sample
            col_min = int(round((grid.sample_grid_bounds[0]-grid.total_bounds.xmin)/grid._search_spacing,0))
            row_min = int(round((grid.sample_grid_bounds[1]-grid.total_bounds.ymin)/grid._search_spacing,0))
            col_max = int(round((grid.sample_grid_bounds[2]-grid.total_bounds.xmin)/grid._search_spacing-1,0))
            row_max = int(round((grid.sample_grid_bounds[3]-grid.total_bounds.ymin)/grid._search_spacing-1,0))
            # row_min, row_max = min([row for row,col in grid.cells_rndm_sample]), max([row for row,col in grid.cells_rndm_sample])
            # col_min, col_max = min([col for row,col in grid.cells_rndm_sample]), max([col for row,col in grid.cells_rndm_sample])
            n_rows_x = row_max - row_min + 1
            n_cols_x = col_max - col_min + 1
            if type(cells_rndm_sample) == bool:
                X = _np_zeros((n_rows_x, n_cols_x), dtype=bool)
                if not cells_rndm_sample:
                    X[:] = True
            else:
                X = _np_zeros((n_rows_x, n_cols_x), dtype=bool)
                X[:] = True
                for (row, col) in cells_rndm_sample:
                    ri = row_max - row
                    ci = col - col_min
                    if 0 <= ri < n_rows_x and 0 <= ci < n_cols_x:
                        X[ri, ci] = False
            cmap_binary = _plt_ListedColormap([sample_area_color, non_valid_area_color])
            extent = [grid.sample_grid_bounds[0],grid.sample_grid_bounds[2],grid.sample_grid_bounds[1],grid.sample_grid_bounds[3]]
            p = ax.imshow(X=X, interpolation='none', cmap=cmap_binary, extent=extent)#, To-Do the extent is imprecise as it does not cover the full grid only its points
            non_valid_patch = _plt_Patch(facecolor=non_valid_area_color, label='Non-valid area', edgecolor=non_valid_area_color)
            sample_patch = _plt_Patch(facecolor=sample_area_color, label='Sample area', edgecolor=sample_area_color)
            ax.legend(handles=[non_valid_patch, sample_patch], loc='best',
                      handlelength=0.8, handleheight=0.8, fontsize=8, borderpad=0.5)
            plot_polygon(poly=non_valid_area, ax=ax, facecolor=non_valid_area_color, edgecolor='black', linewidth=sample_area_lw)
        # ADD DISTRIBUTION PLOT
        c = src.pts[colname]
        scatter_kwargs = {k: v for k, v in plot_kwargs.items() if k not in ('vmin', 'vmax', 'norm')}

        # Detect discrete integer columns (e.g. cluster IDs): ≤20 unique integer values
        _unique_vals = c.dropna().unique()
        _is_discrete = (
            len(_unique_vals) <= 20 and
            all(float(v) == int(v) for v in _unique_vals if v == v)
        )

        if _is_discrete:
            _sorted = sorted(int(v) for v in _unique_vals)
            _n = len(_sorted)
            _qual_cmap = _plt_get_cmap('tab10' if _n <= 10 else 'tab20')
            _colors = ['#cccccc'] + [_qual_cmap(i / max(_n - 2, 1)) for i in range(_n - 1)]
            _disc_cmap = _plt_ListedColormap(_colors)

            _bounds = [v - 0.5 for v in _sorted] + [_sorted[-1] + 0.5]
            _disc_norm = _plt_BoundaryNorm(_bounds, _disc_cmap.N)
            scttr = ax.scatter(x=xs, y=ys, c=c, norm=_disc_norm, cmap=_disc_cmap, rasterized=True, linewidths=0.3, **scatter_kwargs)
            cbar = fig.colorbar(scttr, ax=ax, fraction=0.046, pad=0.04)
            cbar.set_ticks(_sorted)
            cbar.set_ticklabels([str(v) for v in _sorted])
        else:
            vmin = plot_kwargs.get('vmin', c.min())
            vmax = plot_kwargs.get('vmax', c.max())
            norm = plot_kwargs.get('norm', _plt_LogNorm(vmin=vmin, vmax=vmax) if vmin > 0 else _plt_Normalize(vmin=vmin, vmax=vmax))
            scttr = ax.scatter(x=xs, y=ys, c=c, norm=norm, cmap=cmap, rasterized=True, linewidths=0.3, **scatter_kwargs)
            fig.colorbar(scttr, ax=ax, fraction=0.046, pad=0.04)
        plot_polygon(poly=non_valid_area, ax=ax, facecolor="none", edgecolor='black', linewidth=sample_area_lw)
        is_last = (i == colnames.size - 1)
        set_map_frame(ax=ax, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                      r=r if is_last else None)

    if not fig is None:
        if filename:
            fig.savefig(filename, **save_kwargs)
        if not show:
            _plt_close(fig)
    return fig
    #
#


from numpy import (
    array as _np_array,
    zeros as _np_zeros,
    unique as _np_unique,
    spacing as _np_spacing,
    linspace as _np_linspace,
)
from matplotlib.pyplot import subplots as _plt_subplots, colorbar as _plt_colorbar
from matplotlib.pyplot import get_cmap as _plt_get_cmap
from matplotlib.pyplot import close as _plt_close
from matplotlib.patches import (Rectangle as _plt_Rectangle, Polygon as _plt_Polygon, Circle as _plt_Circle)
from matplotlib.collections import PatchCollection as _plt_PatchCollection
from matplotlib.colors import LogNorm as _plt_LogNorm, Normalize as _plt_Normalize
from mpl_toolkits.axes_grid1 import make_axes_locatable as _make_axes_locatable
from aabpl.illustrations.plot_utils import truncate_colormap, map_2D_to_rgb, get_2D_rgb_colobar_kwargs, add_color_bar_ax, set_map_frame
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars

class GridPlots(object):
    """
    Plot methods attached to a Grid object after running ``radius_search``,
    ``detect_cluster_pts``, or ``detect_cluster_cells``.

    Access via ``grid.plot.<method>()``.

    All methods share these common parameters
    ----------------------------------------
    filename : str
        File path to save the figure (e.g. ``'output/my_plot.png'``).
        Leave empty (default ``''``) to skip saving.
    show : bool
        Whether to display the figure inline (Jupyter) or on screen.
        Set ``False`` to save without showing (default ``True``).
    fig / ax / axs
        Pass an existing matplotlib Figure and Axes to draw into them
        instead of creating new ones.
    display_dpi : int
        Resolution used when rendering the figure on screen / in Jupyter
        (default ``100``). Lower values render faster. Has no effect when
        ``show=False``.
    save_kwargs : dict
        Kwargs forwarded to ``fig.savefig()``.
        Defaults: ``{'dpi': 300, 'bbox_inches': 'tight'}``.
        Override dpi for saved files independently:
        ``save_kwargs={'dpi': 150, 'format': 'svg'}``.
    **plot_kwargs
        Additional keyword arguments forwarded to the underlying
        matplotlib call. Supported keys vary per method — see each
        method's docstring.

    Methods
    -------
    clusters       : Detected cluster polygons overlaid on aggregated cell values.
    cell_aggregates: Aggregated indicator values per grid cell (heatmap).
    cluster_vars   : Scatter of source points coloured by aggregated value.
    rand_dist      : Comparison of random vs. observed radius-sum distribution.
    cluster_pts    : Source points coloured by cluster membership.
    grid_ids       : Grid cell row/col indices and point counts (diagnostic).

    Examples
    --------
    fig = grid.plot.clusters(filename='output/clusters.png')
    fig = grid.plot.clusters(filename='output/clusters.png', show=False)
    fig = grid.plot.clusters(figsize=(20, 10), cmap='Blues')
    """
    def __init__(self, grid):
        self.grid = grid

    def vars(self, colnames=None, filename='', **plot_kwargs):
        from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
        import numpy as np
        if colnames is None:
            _sc = getattr(self.grid, '_search_class', None) or self.grid.search
            colnames = np.array(_sc.target.c)
        _r = getattr(self.grid, '_r', None)
        return create_plots_for_vars(grid=self.grid, colnames=colnames, filename=filename, plot_kwargs=plot_kwargs, r=_r)

    def cell_aggregates(
        self,
        filename:str='',
        fig=None,
        axs=None,
        show:bool=True,
        display_dpi:int=100,
        save_kwargs:dict={},
        **plot_kwargs
    ):
        """
        Plot the aggregated indicator value per grid cell as a heatmap.
        One subplot is created per column in ``c``.

        Parameters
        ----------
        filename : str
            Save path. Empty string skips saving (default ``''``).
        show : bool
            Display the figure (default ``True``).
        display_dpi : int
            Resolution for inline/screen display (default ``100``).
            Use ``save_kwargs={'dpi': 300}`` to control saved-file resolution.
        fig, axs
            Existing Figure / Axes array to draw into.
        save_kwargs : dict
            Forwarded to ``fig.savefig()``.
        **plot_kwargs
            figsize : tuple, default ``(12, 10)``
            cmap : str, default ``'Reds'``

        Returns
        -------
        matplotlib.figure.Figure
        """
        save_kwargs = {'dpi': 300, 'bbox_inches': 'tight', **save_kwargs}
        figsize = plot_kwargs.pop('figsize', (12, 10))
        cmap_name = plot_kwargs.pop('cmap', 'Reds')
        _sc = getattr(self.grid, '_search_class', None) or self.grid.search
        if axs is None:
            fig, axs = _plt_subplots(ncols=len(_sc.target.c), figsize=figsize, dpi=display_dpi)

        # Output grid + cached per-output-cell aggregates are built lazily here.
        self.grid.update_spacing()
        # Draw one coloured square per NON-EMPTY output cell (from the cached
        # grid.output_id_to_sums) as a PatchCollection — no dense full-grid array,
        # so memory scales with populated cells, not the bounding-box area. The axes
        # still span the full output extent, so the plot looks the same as a raster.
        sx, sy = self.grid.output_spacing, self.grid.output_spacing_y
        ox, oy = self.grid.output_x_steps[0], self.grid.output_y_steps[0]
        extent = [self.grid.output_x_steps[0], self.grid.output_x_steps[-1],
                  self.grid.output_y_steps[-1], self.grid.output_y_steps[0]]
        out_sums = self.grid.cell_aggregates
        for i,column in enumerate(_sc.target.c):
            ax = axs if len(_sc.target.c)==1 else axs.flat[i]
            rects, vals = [], []
            for (row, col), v in out_sums.items():
                val = v[i]
                if val == 0:
                    continue
                rects.append(_plt_Rectangle((ox + int(col) * sx, oy + int(row) * sy), sx, sy))
                vals.append(val)
            if not vals:
                continue
            vals = _np_array(vals)
            vmin = vals[vals != 0].min()
            vmax = vals.max()
            norm = _plt_LogNorm(vmin=vmin,vmax=vmax,clip=False) if vmin>0 else _plt_Normalize(vmin=vmin,vmax=vmax,clip=False)
            cmap = truncate_colormap(_plt_get_cmap(cmap_name), 0.3, 1)
            cmap.set_under('#ffffff00')
            cmap.set_bad('#ffffff00')
            pc = _plt_PatchCollection(rects, cmap=cmap, norm=norm)
            pc.set_array(vals)
            ax.add_collection(pc)
            cb = _plt_colorbar(pc, cax=add_color_bar_ax(fig,ax))
            ax.set_xlabel('x/lon')
            ax.set_ylabel('y/lat')
            ax.title.set_text('Aggregated value per cell for '+str(column))
            set_map_frame(ax=ax, xmin=extent[0], xmax=extent[1], ymin=extent[3], ymax=extent[2], r=getattr(self.grid, '_r', None))
        if not fig is None:
            if filename:
                fig.savefig(filename, **save_kwargs)
            if not show:
                _plt_close(fig)
        return fig

    #

    def clusters(self, filename:str='', fig=None, axs=None, show:bool=True, display_dpi:int=100, save_kwargs:dict={}, **plot_kwargs):
        """
        Plot detected cluster polygons overlaid on aggregated cell values.
        One subplot per column in ``c``. Cluster outlines are drawn in red
        with hatching, labelled by cluster id.

        Parameters
        ----------
        filename : str
            Save path. Empty string skips saving (default ``''``).
        show : bool
            Display the figure (default ``True``).
        display_dpi : int
            Resolution for inline/screen display (default ``100``).
            Use ``save_kwargs={'dpi': 300}`` to control saved-file resolution.
        fig, axs
            Existing Figure / Axes array to draw into.
        save_kwargs : dict
            Forwarded to ``fig.savefig()``.
        **plot_kwargs
            figsize : tuple, default ``(10, 10 * n_indicators)``
            cmap : str, default ``'binary'``

        Returns
        -------
        matplotlib.figure.Figure

        Examples
        --------
        fig = grid.plot.clusters()
        fig = grid.plot.clusters(filename='clusters.png', show=False)
        fig = grid.plot.clusters(figsize=(20, 15), cmap='Greys')
        """
        if len(self.grid.clustering.by_column)==0:
            print("No clustering performed. Run detect_cell_clusters or grid.create_clusters first.")
            return
        save_kwargs = {'dpi': 300, 'bbox_inches': 'tight', **save_kwargs}
        _sc2 = getattr(self.grid, '_search_class', None) or self.grid.search
        n = len(_sc2.target.c)
        figsize = plot_kwargs.pop('figsize', (10, 10 * n))
        cmap_name = plot_kwargs.pop('cmap', 'binary')
        if axs is None:
            fig, axs = _plt_subplots(ncols=n, figsize=figsize, dpi=display_dpi)

        # Output grid + cached per-output-cell aggregates built lazily here; the
        # background matches grid.plot.cell_aggregates. Cluster outlines are drawn in
        # coordinate space below and are unaffected by the grid choice.
        self.grid.update_spacing()
        sx, sy = self.grid.output_spacing, self.grid.output_spacing_y
        ox, oy = self.grid.output_x_steps[0], self.grid.output_y_steps[0]
        extent = [self.grid.output_x_steps[0], self.grid.output_x_steps[-1],
                  self.grid.output_y_steps[-1], self.grid.output_y_steps[0]]
        out_sums = self.grid.cell_aggregates

        for i, (cluster_column, clusters_for_column) in enumerate(self.grid.clustering.by_column.items()):
            ax = axs.flat[i] if n > 1 else axs
            ax.set_xlabel('x/lon '+str(self.grid.proj_crs))
            ax.set_ylabel('y/lat '+str(self.grid.proj_crs))
            clusters = clusters_for_column.clusters
            ax.title.set_text(str(len(clusters))+' cluster'+ ('s' if len(clusters)!=1 else '') +' for '+str(cluster_column))

            # one coloured square per non-empty output cell (sparse, no dense raster)
            rects, vals = [], []
            for (row, col), v in out_sums.items():
                val = v[i]
                if val == 0:
                    continue
                rects.append(_plt_Rectangle((ox + int(col) * sx, oy + int(row) * sy), sx, sy))
                vals.append(val)
            cmap = _plt_get_cmap(cmap_name)
            cmap = truncate_colormap(cmap, 0.1, 1)
            cmap.set_under('#fff0')
            if vals:
                vals = _np_array(vals)
                vmin, vmax = vals[vals != 0].min(), vals.max()
                norm = _plt_LogNorm(vmin=vmin,vmax=vmax,clip=False) if vmin>0 else _plt_Normalize(vmin=vmin,vmax=vmax,clip=False)
                pc = _plt_PatchCollection(rects, cmap=cmap, norm=norm)
                pc.set_array(vals)
                ax.add_collection(pc)
            for cluster in clusters:
                geoms = [cluster.geometry] if hasattr(cluster.geometry, 'exterior') else cluster.geometry.geoms
                for geom in geoms:
                    ax.add_patch(_plt_Polygon(xy=geom.exterior.coords, hatch='////', facecolor='#f000', edgecolor='#f00'))
                ax.annotate(cluster.id, xy=cluster.centroid, fontsize=15, weight='bold', color='red')

            _r = getattr(self.grid, '_r', None)
            set_map_frame(ax=ax, xmin=self.grid._search_x_steps.min(), xmax=self.grid._search_x_steps.max(),
                          ymin=self.grid._search_y_steps.min(), ymax=self.grid._search_y_steps.max(), r=_r)
            if len(vals) > 0:
                _divider = _make_axes_locatable(ax)
                _cax = _divider.append_axes("right", size="3%", pad=0.05)
                _plt_colorbar(pc, cax=_cax)

        if not fig is None:
            if filename:
                fig.savefig(filename, **save_kwargs)
            if not show:
                _plt_close(fig)
        return fig

    def cluster_vars(
            self,
            filename:str='',
            save_kwargs:dict={},
            show:bool=True,
            display_dpi:int=100,
            **plot_kwargs,
        ):
        """
        Scatter plot of source points coloured by their aggregated radius-sum
        value, alongside the cluster membership column. One subplot per
        indicator column.

        Parameters
        ----------
        filename : str
            Save path. Empty string skips saving (default ``''``).
        show : bool
            Display the figure (default ``True``).
        display_dpi : int
            Resolution for inline/screen display (default ``100``).
            Use ``save_kwargs={'dpi': 300}`` to control saved-file resolution.
        save_kwargs : dict
            Forwarded to ``fig.savefig()``.
        **plot_kwargs
            figsize : tuple, default ``(10 * n_cols, 8 * n_rows)``
            cmap : str, default ``'Reds'``
            s : float  — marker size (default auto-scaled to figsize)
            vmin, vmax : float — colormap range
            sample_area_colors : list[str, str]
                Two hex colours for [valid area, excluded area]
                (default ``['#ffffff', '#bedbe6']``)
            sample_area_linewidth : float
                Border linewidth of the valid-area polygon (default ``0.5``).

        Returns
        -------
        matplotlib.figure.Figure

        Examples
        --------
        fig = grid.plot.cluster_vars()
        fig = grid.plot.cluster_vars(filename='vars.png', show=False, cmap='viridis')
        fig = grid.plot.cluster_vars(figsize=(20, 10), s=2)
        """
        return create_plots_for_vars(
            grid=self.grid,
            colnames=_np_array([(_sc3 := (getattr(self.grid, '_search_class', None) or self.grid.search)).target.c, _sc3.source.aggregate_columns]),
            filename=filename,
            show=show,
            save_kwargs=save_kwargs,
            plot_kwargs=plot_kwargs,
            display_dpi=display_dpi,
        )
    #

    def grid_ids(self, fig=None, ax=None, filename:str='', show:bool=True, display_dpi:int=100, save_kwargs:dict={}, **plot_kwargs):
        """
        Diagnostic plot showing grid cell coordinates, row/column indices,
        and point counts per cell (three side-by-side panels).

        Parameters
        ----------
        filename : str
            Save path. Empty string skips saving (default ``''``).
        show : bool
            Display the figure (default ``True``).
        display_dpi : int
            Resolution for inline/screen display (default ``100``).
            Use ``save_kwargs={'dpi': 300}`` to control saved-file resolution.
        fig, ax
            Existing Figure / Axes to draw into.
        save_kwargs : dict
            Forwarded to ``fig.savefig()``.
        **plot_kwargs
            figsize : tuple, default ``(15, 10)``

        Returns
        -------
        matplotlib.figure.Figure
        """
        save_kwargs = {'dpi': 300, 'bbox_inches': 'tight', **save_kwargs}
        figsize = plot_kwargs.pop('figsize', (15, 10))
        if ax is None:
            fig, ax = _plt_subplots(ncols=3, figsize=figsize, dpi=display_dpi)
        imshow_kwargs = {
            'xmin':self.grid._search_x_steps.min(),
            'ymin':self.grid._search_y_steps.min(),
            'xmax':self.grid._search_x_steps.max(),
            'ymax':self.grid._search_y_steps.max(),
        }
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]
        X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.grid._search_x_steps[:-1]] for y in reversed(self.grid._search_y_steps[:-1])])
        # ax.flat[0].imshow(X=X, interpolation='none', extent=extent)
        # ax.flat[0].pcolormesh([self.grid._search_x_steps, self.grid._search_y_steps], X)
        # ax.flat[0].pcolormesh(X, edgecolor="black", linewidth=.1/max([len(self.grid._search_col_ids), len(self.grid._search_row_ids)]))
        ax.flat[0].imshow(X=X, interpolation='none', extent=extent)
        # ax.flat[0].set_aspect(2)
        colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
        cb = _plt_colorbar(**colorbar_kwargs[2], ax=ax.flat[0])
        cb.ax.set_xlabel("diagonal")
        cb = _plt_colorbar(**colorbar_kwargs[0], ax=ax.flat[0])
        cb.ax.set_xlabel("x/lon")
        cb = _plt_colorbar(**colorbar_kwargs[1], ax=ax.flat[0])
        cb.ax.set_xlabel("y/lat") 
        ax.flat[0].set_xlabel('x/lon') 
        ax.flat[0].set_ylabel('y/lat') 
        ax.flat[0].title.set_text("Grid lat / lon coordinates")

        imshow_kwargs = {
            'xmin':self.grid._search_col_ids.min(),
            'ymin':self.grid._search_row_ids.min(),
            'xmax':self.grid._search_col_ids.max(),
            'ymax':self.grid._search_row_ids.max(),
        }
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]

        X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.grid._search_col_ids] for y in reversed(self.grid._search_row_ids)])
        ax.flat[1].imshow(X=X, interpolation='none', extent=extent)
        # ax.flat[1].set_aspect(2)
        colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
        # cb = _plt_colorbar(**colorbar_kwargs[2], ax=ax.flat[1])
        cb = _plt_colorbar(**colorbar_kwargs[2], cax=add_color_bar_ax(fig,ax.flat[1]))
        cb.ax.set_xlabel("diagonal")
        # cb = _plt_colorbar(**colorbar_kwargs[0], ax=ax.flat[1])
        cb = _plt_colorbar(**colorbar_kwargs[0], cax=add_color_bar_ax(fig,ax.flat[1]))
        cb.ax.set_xlabel("col nr")
        cb = _plt_colorbar(**colorbar_kwargs[1], ax=ax.flat[1])
        cb = _plt_colorbar(**colorbar_kwargs[1], cax=add_color_bar_ax(fig,ax.flat[1]))
        cb.ax.set_xlabel("row nr") 
        ax.flat[1].set_xlabel('row nr') 
        ax.flat[1].set_ylabel('col nr') 
        ax.flat[1].title.set_text("Grid row / col indices")
        
        from aabpl.radius_search.point_grid_assignment import cell_count as _cell_count
        X = _np_array([[_cell_count(self.grid, row_id, col_id) for col_id in self.grid._search_col_ids] for row_id in reversed(self.grid._search_row_ids)])
        # p = ax.flat[2].pcolormesh(X, cmap='Reds')
        p = ax.flat[2].imshow(X=X, interpolation='none', extent=extent, cmap='Reds')
        _plt_colorbar(p, cax=add_color_bar_ax(fig,ax.flat[2]))
        ax.flat[2].set_xlabel('row nr') 
        ax.flat[2].set_ylabel('col nr') 
        if not fig is None:
            if filename:
                fig.savefig(filename, **save_kwargs)
            if not show:
                _plt_close(fig)
        return fig
    #
#


from types import SimpleNamespace


class _DummyEdge:
    def __init__(self, xy1, xy2):
        self.vtx1 = SimpleNamespace(xy=xy1)
        self.vtx2 = SimpleNamespace(xy=xy2)


def test_split_region_validation_rejects_degenerate_chain():
    from aabpl.radius_search.region_classes import _region_is_valid_for_split

    region = SimpleNamespace(
        edges=[_DummyEdge((0.0, 0.0), (1.0, 0.0)), _DummyEdge((1.0, 0.0), (1.0, 1.0))],
        vertices=[
            SimpleNamespace(xy=(0.0, 0.0)),
            SimpleNamespace(xy=(1.0, 0.0)),
            SimpleNamespace(xy=(1.0, 1.0)),
        ],
        is_closed=False,
    )

    assert _region_is_valid_for_split(region) is False
